/*
 * tclXutil.c
 *
 * Utility functions for Extended Tcl.
 *-----------------------------------------------------------------------------
 * Copyright 1991-1994 Karl Lehenbauer and Mark Diekhans.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted, provided
 * that the above copyright notice appear in all copies.  Karl Lehenbauer and
 * Mark Diekhans make no representations about the suitability of this
 * software for any purpose.  It is provided "as is" without express or
 * implied warranty.
 *-----------------------------------------------------------------------------
 * $Id: tclXutil.c,v 1.2 1996/05/13 21:24:20 flisakow Exp $
 *-----------------------------------------------------------------------------
 */

/*  #include "tclExtdInt.h"  */

#include <stdlib.h>
#include <tcl.h>
#include <tk.h>

#ifndef _tolower
#  define _tolower tolower
#  define _toupper toupper
#endif

#ifndef TRUE
#    define TRUE   (1)
#    define FALSE  (0)
#endif

extern int errno;

#define ckstrdup(sourceStr) \
  (strcpy (malloc (strlen (sourceStr) + 1), sourceStr))

/*
 * Used to return argument messages by most commands.
 */
static char *tclXWrongArgs = "wrong # args: ";


/*
 *-----------------------------------------------------------------------------
 *
 * Tcl_SetupFileEntry --
 *
 * Set up an entry in the Tcl file table for a file number, including the stdio
 * FILE structure.
 *
 * Parameters:
 *   o interp (I) - Current interpreter.
 *   o fileNum (I) - File number to set up the entry for.
 *   o permissions (I) - Flags consisting of TCL_FILE_READABLE,
 *     TCL_FILE_WRITABLE.
 * Returns:
 *   A pointer to the FILE structure for the file, or NULL if an error
 * occured.
 *-----------------------------------------------------------------------------
 */
FILE *
Tcl_SetupFileEntry (interp, fileNum, permissions)
    Tcl_Interp *interp;
    int         fileNum;
    int         permissions;
{
    /*  Interp   *iPtr = (Interp *) interp;  */
    char     *mode;
    FILE     *filePtr;

    /*
     * Set up a stdio FILE control block for the new file.
     */
    if (permissions & TCL_FILE_WRITABLE) {
        if (permissions & TCL_FILE_READABLE)
            mode = "r+";
        else
            mode = "w";
    } else {
        mode = "r";
    }

    filePtr = fdopen (fileNum, mode);
    if (filePtr == NULL) {
        interp->result = Tcl_PosixError (interp);
        return NULL;
    }
    
    Tcl_EnterFile (interp, filePtr, permissions);

    return filePtr;
}

/*
 *-----------------------------------------------------------------------------
 *
 * Tcl_CloseForError --
 *
 *   Close a file number on error.  If the file is in the Tcl file table, clean
 * it up too. The variable errno, and interp->result and the errorCode variable
 * will be saved and not lost.
 *
 * Parameters:
 *   o interp (I) - Current interpreter.
 *   o fileNum (I) - File number to close.
 *-----------------------------------------------------------------------------
 */
void
Tcl_CloseForError (interp, fileNum)
    Tcl_Interp *interp;
    int         fileNum;
{
    static char *ERROR_CODE = "errorCode";
    int          saveErrNo = errno;
    char        *saveResult, *errorCode, *saveErrorCode, *argv [2], buf [32];

    saveResult = ckstrdup (interp->result);

    errorCode = Tcl_GetVar (interp, ERROR_CODE, TCL_GLOBAL_ONLY);
    if (errorCode != NULL)
        saveErrorCode = ckstrdup (errorCode);
    else
        saveErrorCode = NULL;

    sprintf (buf, "file%d", fileNum);

    argv [0] = "close";
    argv [1] = buf;
    Tcl_CloseCmd (NULL, interp, 2, argv);
    Tcl_ResetResult (interp);

    if (saveErrorCode != NULL) {
        Tcl_SetVar (interp, ERROR_CODE, saveErrorCode, TCL_GLOBAL_ONLY);
        free (saveErrorCode);
    }
    Tcl_SetResult (interp, saveResult, TCL_VOLATILE);
    free (saveResult);

    close (fileNum);  /* In case Tcl didn't have it open */
    
    errno = saveErrNo;
}
     
