//********************************
// HashJoin.C
//
// Ravi Murthy
//********************************

#ifdef    WITH_PERSISTENCE
#include "HashJoin.h"

//** Forward declaration

void myunify(Cor_ArgList& arg1, Cor_ArgList& arg2,
            Cor_BindEnv* env1, Cor_BindEnv *env2);

//***************************************************
// The constructor builds a hash table using the iterator on the
// the relation specified.
Cor_HashJoin::Cor_HashJoin(Cor_TupleIterator* inner, Cor_JoinAttrs *attr)
    :inner_arg_list(inner->arg_list)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    // Store joinattrs in private member
    attrs = attr;

    // Initialize some private members
    save_bindenv = probe_bindenv = NULL;
    save_elemptr = NULL;
    probe_done = 1;

    // Initialize hash table
    for(int i = 0; i < TABLE_SIZE; i++)
        table[i] = NULL;


    // Build hash table for the inner relation
    // Hash on the join attributes

    int hashval;
    Cor_BindEnv *env;

    env = inner->get_next();
    while (!(inner->no_match())) { // found tuple
        Cor_BindEnv *envptr = env->dup();
        hashval = attr->hash_value(envptr) % TABLE_SIZE;
        Cor_ELEM *elemptr = new Cor_ELEM;
        elemptr->tuple = envptr;
        elemptr->next = table[hashval];
        table[hashval] = elemptr;
        env = inner->get_next();
    }
}

// Destructor
Cor_HashJoin::~Cor_HashJoin()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    // Free all entries of the hash table.

    Cor_ELEM *elemptr;
    Cor_ELEM *nextelemptr;

    for (int i = 0; i < TABLE_SIZE; i++) {
        elemptr = table[i];
        while (elemptr != NULL) {
            nextelemptr = elemptr->next;
            delete elemptr;
            elemptr = nextelemptr;
        }
    }
}

// *************************************************************************
// This function uses the outer iterator to get the next tuple. (actually
// a bindenv). Then, it computes the hash value. We look up the hash table
// at this value and compare each of the stored bindenv with this one.
// If any of them match, the resulting bindenv is returned. 
Cor_Tuple* Cor_HashJoin::get_next_tuple(Cor_TupleIterator *outer)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_ELEM *elemptr = NULL;
    Cor_Tuple *outertuple;
    int hashval;

    while (1) {
        if (!(probe_done)) {
            // The bucket previously being probed should be continued
            // SO, no outer next tuple to be got.
            probe_bindenv = save_bindenv->dup();
            elemptr = save_elemptr;
        } else {
            // Have to read in the next outer tuple
            outertuple = outer->get_next_tuple();
            if (outer->no_match())  // outer has run out of tuples
                return NULL;
            
            // hash the tuple
            hashval = attrs->hash_value(outer->bindenv) % TABLE_SIZE;
            
            // Probe the hash table
            elemptr = table[hashval];
    
            probe_bindenv = outer->bindenv->dup();
            save_bindenv = outer->bindenv->dup();
        }
    
        while (elemptr != NULL) {
            if (attrs->compare(probe_bindenv, elemptr->tuple,
                    outer->arg_list, inner_arg_list)) {
                // found join
                outer->reset_no_match();
    
                myunify(outer->arg_list, inner_arg_list,
                            outer->bindenv, probe_bindenv);
                save_elemptr = elemptr->next;
                probe_done = 0;
                return NULL;
            }
            // check next
            elemptr = elemptr->next;
        }
    
        // Probe failed
        // Check next outer tuple
        probe_done = 1;
        probe_bindenv = save_bindenv = NULL;
    }
}

// *************************************************************************
// This function unifies the arguments from arg1 and arg2 from the bindenv
// env2 to env1.
void myunify(Cor_ArgList& arg1, Cor_ArgList& arg2,
        Cor_BindEnv* env1, Cor_BindEnv *env2)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int i;
    Cor_Term term;
    Cor_CArg *var1, *var2;

    for (i = 0; i < arg1.count(); i++) {
        var1 = &(arg1[i]);
        term = env2->lookup(var1->var_num());
        Cor_unify_binding(env1, var1->var_num(), term);
    }

    for (i = 0; i < arg2.count(); i++) {
        var2 = &(arg2[i]);
        term = env2->lookup(var2->var_num());
        Cor_unify_binding(env1, var2->var_num(), term);
    }
}

// *************************************************************************
#endif // WITH_PERSISTENCE
