// *************************************************************************
//    CoralInit.C
// *************************************************************************

#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <signal.h>

#include <new.h>

#if defined(SOLAR_386) || defined(SOLAR)
#    include <sys/resource.h>
#    include <sys/procfs.h>
#endif

#include "CoralInit.h"

#include "Globals.h"
#include "Database.h"
#include "Execute.h"
#include "CoralError.h"
#include "CoralBreak.h"
#include "Util.h"
#include "Relation.h"
#include "InitBuiltins.h"
#include "BuiltinSymbols.h"
#include "SMUtil.h"
#include "LocalConnectArray.h"
#include "Suspension.h"
#include "Compile.h"
#include "Entity.h"
#include "StrRefTable.h"
#include "RdbInit.h"
#include "Debug.h"

static char what_version[] = "@(#) Coral Version 1.5.1";

// *************************************************************************

extern void clean_up_static_terms();

static void exceptionHandler(int recvd_signal);
static void freeStoreException();

#ifndef DEBUG
static void brokenPipe(int recvd_signal);
#endif

// *************************************************************************

extern void sm_init();

char coral_name[200];        // full pathname of the executing program

int Cor_startMemUsage;

#if defined(SUN) || defined(LINUX) || defined(HPUX)
    struct rusage Cor_startRUsage;
    struct rusage Cor_SystemUsage;
#endif

#if defined(SOLAR_386) || defined(SOLAR)
    //struct prusage Cor_SystemPrUsage;
#ifdef RUSAGE_SELF
    struct rusage  Cor_SystemUsage;
#endif // RUSAGE_SELF
#endif

int Cor_beginMem;

int use_rdb = 0;
int use_exodus = 0;

Cor_DatabaseStruct *Cor_ExodusDB = NULL;

// *************************************************************************
//    System specific defines & includes

#ifdef HPUX
    typedef void (*COR_SIGNAL_HANDLER) (int);
#    include <sys/syscall.h>
#    define getrusage(a, b)  syscall(SYS_GETRUSAGE, a, b)
#else
#    ifdef __GNUC__
#        define COR_SIGNAL_HANDLER
#    else
#    define COR_SIGNAL_HANDLER SIG_PF
#    endif
#endif

// *************************************************************************
void Cor_new_globals()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    // Get the global variables for Coral

    // This function will almost certainly be called multiple times,
    // so you must always defend against that.

    // String reference count table
    if (!Cor_string_ref_table) {
        Cor_string_ref_table = new Cor_StrRefTable();

        // init builtin symbols
        Cor_init_builtin_symbols();
    }

        // Open standard in/out/err as CoralStreams
    if (cout_strm == NULL) {
        cin_strm  = new CoralStream(cin.rdbuf(), 0);
        cout_strm = new CoralStream(cout.rdbuf(), 0);
        cerr_strm = new CoralStream(cerr.rdbuf(), 0);
    }

#ifdef    REUSE_MEMORY
    if (Coral_Free == NULL) {
        Coral_Free = new Cor_Entity_Array();
    }
#endif

    if (Cor_parserStruct == NULL) {
        Cor_CurTrail     = new Cor_Trail(200);
        Cor_deflt_exEnv  = new Cor_ExecutionEnv();
        Cor_exEnv        = Cor_deflt_exEnv;
        Cor_parserStruct = new Cor_ParserStruct();
        coralConfig      = new CoralConfig();
        Cor_T_Stack      = new Cor_TableStack(COR_MAX_RECURSION_DEPTH);
    }

    if (Cor_BuiltinDB == NULL) {
          // lame - get rid of the "nil" thing?
        Cor_BuiltinDB = new Cor_DatabaseStruct(*Cor_NilSymbol);
        Cor_CurDB     = Cor_BuiltinDB;
    }

    if (Cor_suspension == NULL) {
        Cor_suspension = new Cor_suspensionClass();
        Cor_brk_Info = new Cor_break_info();
        Cor_UnusedEnv = new Cor_BindEnv(0);
        Cor_static_terms = new Cor_Term[3];
        Cor_usable_terms = new Cor_Term[10];

        Cor_TheDontCareArg  = new Cor_CArg(DONT_CARE_ARG);
    }
}

// *************************************************************************
void Cor_delete_globals()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    // Get rid of the global variables for Coral

    clean_up_static_terms();

    delete Cor_HelpRel;

    delete Cor_TheDontCareArg;

    delete [] Cor_static_terms;
    delete [] Cor_usable_terms;
    delete Cor_UnusedEnv;
    delete Cor_brk_Info;

    delete Cor_parserStruct;

    delete Cor_clients;
    delete Cor_suspension;

    delete Cor_BuiltinDB;
    delete Cor_DatabaseTable;

    delete Cor_T_Stack;
    delete coralConfig;
    delete Cor_CurTrail;

    delete Cor_deflt_exEnv;


      // Set it up so init_coral() _could_ be called again
      // (from the C++ Interface)
    Cor_TheDontCareArg  = NULL;
    Cor_UnusedEnv       = NULL;
    Cor_brk_Info        = NULL;
    Cor_parserStruct    = NULL;
    Cor_clients         = NULL;
    Cor_suspension      = NULL;
    Cor_BuiltinDB       = NULL;
    Cor_DatabaseTable   = NULL;
    Cor_T_Stack         = NULL;
    coralConfig         = NULL;
    Cor_CurTrail        = NULL;
    Cor_deflt_exEnv     = NULL;
}

// *************************************************************************
//    Initializes CORAL.  Called from main routine.

void init_coral(char *pgm)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif

      // Sets the values of all global variables
    Cor_new_globals();
    
#    ifndef DEBUG
        // Catch signals that might otherwise cause termination
        for (int i = SIGINT; i < SIGPIPE; i++) {
            if ((i != SIGKILL) && (i != SIGTERM)) {
#                ifdef __GNUC__
                    signal(i, (exceptionHandler));
#                else
                    signal(i, (COR_SIGNAL_HANDLER)(exceptionHandler));
#                endif
            }
        }

          // Catch broken pipes (An Explain that quit)
#       ifdef __GNUC__
            signal(SIGPIPE, (brokenPipe));
#       else
            signal(SIGPIPE, (COR_SIGNAL_HANDLER)(brokenPipe));
#       endif

#    else
        // for debugging purposes, catch only SIGINT (CTRL-C)
#        ifdef __GNUC__
            signal(SIGINT, (exceptionHandler));
#        else
            signal(SIGINT, (COR_SIGNAL_HANDLER)(exceptionHandler));
#        endif
#    endif

    sprintf(coral_pid, "%lu", (unsigned long) getpid());

    // If a ConnectArray hasn't been set up, make it
    // a local connection with an empty stack.
    if (! Cor_clients) {
        Cor_clients = new Cor_LocalConnectArray();
          // Remove stdin from the stack
        Cor_parseStack->abort_file();
    } 

    // Provide exception handler for new()
    set_new_handler(freeStoreException);

    // create the CurrModuleInfo structure
    Cor_CurrModuleInfo = new Cor_ModuleInfo();

    strcpy(coral_name, (pgm ? pgm : "coral"));

    if ((Cor_exEnv->C_exec_mode == COR_DOING_INTERPRET) ||
        (Cor_exEnv->C_exec_mode == COR_DOING_IMPERATIVE))
    {
        // initialize global stats for CPU and memory usage

#        ifdef SUN
            getrusage(RUSAGE_SELF, &Cor_startRUsage);
#        endif

#        ifdef LINUX
            getrusage(RUSAGE_SELF, &Cor_startRUsage);
#        endif


        //Cor_startMemUsage = (int)sbrk(0);

#        ifdef WITH_PERSISTENCE

         if (use_exodus) {

            if (Cor_exEnv->C_connect_exodus_default) {

                  // Create exodus workspace
                Cor_ExodusDB = new Cor_DatabaseStruct(Cor_CArg(COR_EXODUS_WS));
                Cor_CurDB = Cor_ExodusDB;
                Cor_CurDB->RelationTable = new Cor_SymTable();

                  // Add exodus database to the Database Table
                if (!Cor_addDBStruct(Cor_CurDB)) {
                    CORAL_error(COR_MULTIPLE_DB_ERR, "unable to add exodus DB",
                        "init-coral", __FILE__, __LINE__);
                }
                else 
                    sm_init();
            }
         }
#        endif

        if (use_rdb)
            init_rdb();

        // Remove the stdin.M file
        unlink("stdin.M");

        // Initialize builtin database.
        Cor_BuiltinDB->name = Cor_CArg("builtin_ws");
        Cor_initBuiltins();
        Cor_CurDB = Cor_BuiltinDB;

        // Add builtin database to the Database Table
        if (!Cor_addDBStruct(Cor_CurDB)) {
            CORAL_error(COR_MULTIPLE_DB_ERR, "unable to add builtin DB",
                "init-coral", __FILE__, __LINE__);
            exit(1);
        }

        // Create default database
        Cor_CArg dbarg = Cor_CArg(COR_DEFAULT_DB);
        Cor_DatabaseStruct *db = Cor_createDB(dbarg);
        if (!db) {
            CORAL_error(COR_MULTIPLE_DB_ERR, "unable to add default DB",
                "init-coral", __FILE__, __LINE__);
            exit(1);
        }
        Cor_CurDB = db;

        if (Cor_exEnv->C_read_coralrc_default
               && !(Cor_exEnv->C_exec_mode == COR_DOING_IMPERATIVE)) {

            if (coralConfig->server_mode && !Cor_clients->is_local()) {
                Cor_clients->LocalConnection();

                  // Remove stdin from the stack
                Cor_parseStack->pop();
            }
            
            // Read .coralrc if it exists, else
            // read input from $(HOME)/.coralrc
            if (!Cor_consultFile(".coralrc", ".coralrc")) {
                char *home_path = getenv("HOME");
               
                if (home_path) { 
                    char *tmp_name = new char[strlen(home_path) + 15];
                    strcpy(tmp_name, home_path);
                    strcat(tmp_name, "/.coralrc");
                    Cor_consultFile(tmp_name, tmp_name);
                    delete [] tmp_name;
                }
            }

            if (coralConfig->server_mode && !Cor_clients->is_local()) {

                while (! Cor_parseStack->is_empty()) {
                  Cor_parseStack->parse(cerr_strm, 1,
                    "syntax error -- skipping to next period\n");
                } 
                Cor_clients->DeleteLocalConnection();
            }
        }
    } else {
        Cor_exEnv->C_do_history_default = 0;
    }

#   ifdef SUN
    getrusage(RUSAGE_SELF, &Cor_SystemUsage);
#   endif

#   ifdef LINUX
    getrusage(RUSAGE_SELF, &Cor_SystemUsage);
#   endif

    Cor_beginMem = 0;
}

// *************************************************************************
//  A few extra things when we're being used interactively
// *************************************************************************
void init_coral_interactive()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
      // Load in the help index
    char help_index[1000];
    sprintf(help_index, "%s%s/INDEX.F", Cor_exEnv->coral_path, HELP_PATH);

    if (! (Cor_HelpRel = Cor_ReadRelation(help_index, 3))) 
        *(Cor_exEnv->err_strm) << "Error: Could not find help Index file.\n";
}

// *************************************************************************
//    Called when quitting coral

void exit_coral()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
#   ifdef WITH_PERSISTENCE

      // This shuts down the Exodus connection
    if (use_exodus)
        shutdown_server();

#   endif

#ifdef    DO_PURE
      // We lose at most one, just get it out of Purify's hair.
    Cor_clear_scc_list();
#endif

    Cor_clearDatabases();

    if (use_rdb)
        exit_rdb();

    Cor_exEnv->flush_all_outputs();

    delete Cor_CurrModuleInfo;

    Cor_delete_globals();

    delete Coral_Free;
    Coral_Free = NULL;

    Cor_free_builtin_symbols();

    delete Cor_string_ref_table;
    Cor_string_ref_table = NULL;
}

// *************************************************************************
//    Signal handler

static void exceptionHandler(int recvd_signal)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    CORAL_error(COR_SIGNAL_RECEIVED,NULL,NULL,__FILE__,__LINE__);
    *(Cor_exEnv->err_strm) << "Received signal no: " <<  recvd_signal << "\n";

    //    Code for recovering from abnormal situation
    if (recvd_signal != SIGINT) {
        *(Cor_exEnv->err_strm) << "Cannot recover from received signal\n";
        *(Cor_exEnv->err_strm) << "Initiating shutdown procedures\n";
#        ifdef WITH_PERSISTENCE
        if (use_exodus)
            abortTrans();
#        endif
        // Is something hanging occasional in this call?
        //exit_coral();
        if (use_rdb)
            exit_rdb();
        exit(1);
    }

    if (Cor_exEnv->C_interrupt_raised == 1) {
        *(Cor_exEnv->err_strm) << "Received second interrupt\n";
        *(Cor_exEnv->err_strm) << "Initiating shutdown procedures\n";
#        ifdef WITH_PERSISTENCE
        if (use_exodus)
            abortTrans();
#        endif
        exit_coral();
        exit(1);
    }

    Cor_exEnv->C_interrupt_raised = 1;
}

// *************************************************************************
//    Signal handler for broken pipe (caused by Explain that disconnects)

#ifndef    DEBUG
static void brokenPipe(int recvd_signal)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    // Assume it was the current users explain (pretty safe assumption)

      // Turn off explain and reset associated structures
    Cor_curUser->Cor_exEnv.explain_strm = (CoralStream *) NULL;

    Cor_curUser->Cor_exEnv.GlobalRelationOptions &= ~REL_EXPLAIN;

      // Notify user
    if (Cor_curUser->Cor_exEnv.err_strm)
        *(Cor_exEnv->err_strm)
            << "Explain disconnected - turning explain off.\n";
}
#endif

// *************************************************************************
//    Exception handler for the new() system call

static void freeStoreException()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    CORAL_error(COR_NO_MEMORY,NULL,"freeStoreException",__FILE__,__LINE__);
    Cor_exEnv->out_strm->flush();
    Cor_exEnv->err_strm->flush();
    if (use_rdb)
        exit_rdb();
    exit(1);
}

// *************************************************************************

