// *************************************************************************
//    HashIndex.C
// *************************************************************************

#include "HashIndex.h"

#include "Globals.h"
#include "Execute.h"
#include "Stack.h"
#include "Term.h"
#include "Debug.h"

// *************************************************************************
//  The following is an optimization to let a static environment to be used in
//  reentrant code.  We assume no parallelism however.
// *************************************************************************

static Cor_BindEnv static_env1(MAX_VARS);
static int static_env1_inuse=DONT_USE_STATIC_ENV;

// *************************************************************************

extern int Cor_subsumes_arg_lists(Cor_ArgList& left, Cor_ArgList& right,
    Cor_BindEnv *left_bindings, Cor_BindEnv *right_bindings);
extern int Cor_subsumes_tuple(Cor_Tuple *tuple, Cor_Tuple *newtuple,
        int display);
int Cor_unify_literal(Cor_TupleIterator& iterator, Cor_Tuple *tuple,
        int display);

// *************************************************************************

#define BACKBONESIZE 16

// *************************************************************************
Cor_HashIndex::Cor_HashIndex(Cor_StorageRelation *rel, Cor_BitVector *aset,
        int buckets, double incr_ratio1, double max_occupancy1)
    : Cor_GenericIndex( rel, aset )
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    // for the 0-arity reln, or for the all-free index
    if ((rel->arity() == 0) || (!aset->any())) {
        buckets = 1;            // Ought to be 0, but too much code probably
                                // depends on a non-empty hash table.
        incr_ratio1 = 1.0;
        max_occupancy1 = 1000000.0;
    }

    if (buckets < 0)
        buckets = Cor_exEnv->C_hash_buckets_default;
    if (incr_ratio1 < 0.0)
        incr_ratio1 = Cor_exEnv->C_incr_ratio_default;
    if (max_occupancy1 < 0.0)
        max_occupancy1 = Cor_exEnv->C_max_occupancy_default;


    hash_size  = buckets;
    hash_count = 0;
    incr_ratio = incr_ratio1;
    max_occupancy = max_occupancy1;
    limit = (int) (hash_size * max_occupancy);

    hash_table = new Cor_TupleStack* [hash_size];

    non_mt_bckts = new Cor_IntStack(hash_size); 


#ifdef    SPEED_HACK
    memset(hash_table, 0, sizeof(Cor_TupleStack*) * hash_size);
#else
    int i;
    for (i = 0; i < hash_size; i++)
        hash_table[i] = NULL;
#endif

    var_chain = new Cor_TupleStack(32);
}

// *************************************************************************
Cor_HashIndex::Cor_HashIndex(Cor_StorageRelation *rel, Cor_ArgList *pat,
    int n_var, Cor_ArgList *v_nm, int buckets,
    double incr_ratio1, double max_occupancy1)
    : Cor_GenericIndex(rel, pat, n_var, v_nm)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (rel->arity() == 0) {
        buckets = 1;    // Ought to be 0, but too much code probably depends on 
                          // a non-empty hash table.
    }

    if (buckets < 0)
        buckets = Cor_exEnv->C_hash_buckets_default;
    if (incr_ratio1 < 0.0)
        incr_ratio1 = Cor_exEnv->C_incr_ratio_default;
    if (max_occupancy1 < 0.0)
        max_occupancy1 = Cor_exEnv->C_max_occupancy_default;

    hash_size = buckets;
    hash_count = 0;
    incr_ratio = incr_ratio1;
    max_occupancy = max_occupancy1;
    limit = (int) (hash_size * max_occupancy);

    hash_table = new Cor_TupleStack* [hash_size];

    non_mt_bckts = new Cor_IntStack(hash_size); 

#ifdef    SPEED_HACK
    memset(hash_table, 0, sizeof(Cor_TupleStack*) * hash_size);
#else
    int i;
    for (i = 0; i < hash_size; i++)
        hash_table[i] = NULL;
#endif

    var_chain = new Cor_TupleStack(32);
}

// *************************************************************************
void Cor_HashIndex::reset_index()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (var_chain)
        var_chain->empty();

    hash_count = 0;
    //limit = (int) (hash_size * max_occupancy);


    int j, k, lim = non_mt_bckts->count();

    for (k=0; k < lim; k++) {

        j = non_mt_bckts->get_ith(k);

        if (hash_table[j]) {
            delete hash_table[j];
            hash_table[j] = NULL;
        }
    }

    non_mt_bckts->empty();
}

// *************************************************************************
void Cor_HashIndex::move_all_entries(Cor_GenericIndex *ind)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int j;
    int k = 0;

    if  (index_type() != ind->index_type()) {
        // WARNING - should also check for matching of indices: 
        // With the current implementation the indices will match,
        // but this is something that must be checked.  Insert a call :
        // if (!matches(ind)) error ....
        // It has been left out for efficiency reasons, and because 
        // matches is not fully implemented yet.
            
        *(Cor_exEnv->err_strm)
            << "CORAL:: Internal Error - move_all_entries type "
            << "mismatch between indices.\n";
        return;
    }

    Cor_HashIndex *hindex = (Cor_HashIndex *) ind;

      // Move nodes in the var_chain list
    for (j=0; j < var_chain->count(); j++) {
      if (var_chain->tuples[j])
          if (! hindex->insert_entry(var_chain->hashs[j],
                        var_chain->tuples[j]) )
              var_chain->tuples[k]->delete_self(); 
    }

    var_chain->empty();

    hash_count = 0;
    Cor_TupleStack *bckt;

    // Move nodes in the hash buckets

    int ix = 0;
    int lim = non_mt_bckts->count();
    for (ix=0; ix < lim; ix++) {

        j = non_mt_bckts->get_ith(ix);

        if ((bckt = hash_table[j])) {

            for (k=0; k < bckt->count(); k++) {
              if (bckt->tuples[k])
                  if (! hindex->insert_entry(bckt->hashs[k],bckt->tuples[k])) {
                      bckt->tuples[k]->delete_self(); 
                  }
            }

            delete bckt;
            hash_table[j] = NULL;
        }
    }

    non_mt_bckts->empty();
}

// *************************************************************************
void Cor_HashIndex::show_all_tuples(ostream &out)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
#ifdef    DEBUG

    out << "-----------------------------------\n";
    out << "In show_all_tuples for HashIndex " << this << endl;

    out << "Non-Empty Buckets: " << *non_mt_bckts << endl;

    out << "\nvar_chain contains:\n";

    int j, k, ix = 0;

    for (j=0; j < var_chain->count(); j++) {
        if (var_chain->tuples[j]) {
            out << "Tuple: " << *(var_chain->tuples[j]); 
            if (var_chain->tuples[j]->deleted)
                out << " (deleted).";
            out << endl;
        }
    }

    Cor_TupleStack *bckt;

    int lim = non_mt_bckts->count();
    for (ix=0; ix < lim; ix++) {
        j = non_mt_bckts->get_ith(ix);
        if ((bckt=hash_table[j])) {
            for (k=0; k < bckt->count(); k++) {
                if (bckt->tuples[k]) {
                    out << "Bucket["<< j <<"]: (" << bckt->tuples[k]
                         << ") " << *(bckt->tuples[k]);
                    if (bckt->tuples[k]->deleted)
                        out << " (deleted).";
                    out << endl;
                }
            }
        }
    }

    out << "-----------------------------------\n";
#endif
}

// *************************************************************************
void Cor_HashIndex::delete_all_tuples(int force_del)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int j, k;

#    ifdef DEBUG
        if (Cor_exEnv->dbg_deletes) {
            *(Cor_exEnv->trace_strm) << "Deleting "
                << hash_count << " tuples from index of relation";
            relation->name.print(NULL, Cor_exEnv->trace_strm);
            *(Cor_exEnv->trace_strm) << endl;
        }
#    endif

    // delete nodes in the var_chain list
    for (j=0; j < var_chain->count(); j++) {
        if (var_chain->tuples[j]) {
            var_chain->tuples[j]->delete_self(force_del);
        }
    }
    var_chain->empty();
    
    hash_count = 0;
    Cor_TupleStack *bckt;

    // delete nodes in the hash buckets

    int ix = 0;
    int lim = non_mt_bckts->count();
    for (ix=0; ix < lim; ix++) {
        j = non_mt_bckts->get_ith(ix);

        if (j >= 0) {
            if ((bckt=hash_table[j])) {
                for (k=0; k < bckt->count(); k++) {
                    if (bckt->tuples[k]) {
                        bckt->tuples[k]->delete_self(force_del);
                    }
                }

                delete bckt;
                hash_table[j] = NULL;
            }
        }
    }

    non_mt_bckts->empty();
}

// *************************************************************************
Cor_HashIndex::~Cor_HashIndex()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif

      // Do not force a delete of the actual tuples
    delete_all_tuples(0);

    for (int j=0; j < hash_size; j++)
        delete hash_table[j];

    delete    var_chain;

    delete [] hash_table;

    delete    non_mt_bckts;
}

// *************************************************************************
HashVal Cor_HashIndex::hash(Cor_ArgList *args, Cor_BindEnv *bindenv)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif

    if (args->count() == 0) {
        // Normally this doesn't really mean that arity==0.  Rather, it
        // means that the TupleIterator was created with an ArgList
        // parameter, which is a convention meaning essentially
        // (VAR1, VAR2, ..., VARn).  However, in pathological cases, it
        // might really be a nullary Relation. Do the same thing in either case.
        return VarHashValue;
    }

    HashVal result;

    if (argset != NULL) {
        // the argument form index will be used
        result = args->hash(*argset, bindenv);
        return(result);
    }

    Cor_StackMark stackmark0;

    // else the pattern form index will be used 

    Cor_BindEnv *pat_env;
    if ((static_env1_inuse++ == 0) && (num_vars_pat <= MAX_VARS)) {
        pat_env = &static_env1;
    } else {
        pat_env = new Cor_BindEnv(num_vars_pat);
    }

    // NOTE: a subsumption check alone is insufficient ! In the presence of
    // non-ground facts, the tuple to insert may be more general than the
    // pattern of the index, and unification is needed to ensure that
    // all desired inserts occur --- PRAVEEN

    // Non-ground facts are out for now, so subsumption is fine. -SPF
    //if (Cor_unify_arg_lists(*pattern, *args, pat_env, bindenv)
    //          == COR_U_SUCCEED) {

    if (Cor_subsumes_arg_lists(*pattern, *args, pat_env, bindenv)) {

        Cor_BitVector aset(var_names->count());
        int i;
        for (i = 0; i < var_names->count(); i++)
            aset.set(i);
        
        result = var_names->hash(aset, pat_env);    // hash on all arguments
        stackmark0.pop_to();
        if ((--static_env1_inuse != 0) || (num_vars_pat > MAX_VARS))
            delete pat_env;

        return(result);
    } 

    // else the tuple and the pattern don't match
    stackmark0.pop_to();
    if ((--static_env1_inuse != 0) || (num_vars_pat > MAX_VARS)) {
        delete pat_env;
    }

    return(UnmatchedHashValue);
}

// *************************************************************************
int Cor_HashIndex::insert_tuple( Cor_Tuple *tuple )
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif

    if (tuple->is_deleted()) {
        // tuple has been logically deleted
        return(0);
    }

    HashVal h_val = hash(&(tuple->args()), NULL);

#    ifdef DO_TRACE
    if (Cor_exEnv->dbg_indexing) {
        *(Cor_exEnv->trace_strm) << "HashIndex::insert_tuple: ";
        relation->name.print(NULL, Cor_exEnv->trace_strm);
        *(Cor_exEnv->trace_strm) << "(";
        tuple->printon(Cor_exEnv->trace_strm);
        *(Cor_exEnv->trace_strm) << ") ";

        if (argset != NULL) {
            // argument form index
            *(Cor_exEnv->trace_strm) << "[hash index on ";
            argset->printon(Cor_exEnv->trace_strm);
            *(Cor_exEnv->trace_strm) << " ]\n";
        } else {
            // pattern form index
            *(Cor_exEnv->trace_strm) << "[pattern matched ";
            pattern->print(NULL, Cor_exEnv->trace_strm);
            *(Cor_exEnv->trace_strm) << "] ";
               *(Cor_exEnv->trace_strm) << "[indexed on ";
            var_names->print(NULL, Cor_exEnv->trace_strm);
            *(Cor_exEnv->trace_strm) << " ]\n";
        }
        *(Cor_exEnv->trace_strm) << "        hash value is " <<  h_val << "\n";
    }
#    endif


    if (h_val == VarHashValue) {

        // link it to the beginning of var_chain
        var_chain->push(h_val, tuple);
#        ifdef DO_TRACE
            if (Cor_exEnv->dbg_indexing) {
                *(Cor_exEnv->trace_strm) << "HashIndex::insert_tuple: ";
                *(Cor_exEnv->trace_strm) << "inserting in var_chain\n";
            }
#        endif
    }
    else if (h_val != UnmatchedHashValue) {
        // link it to the beginning of the appropriate hash bucket

        hash_count++;
        int tmp_hash_index = abs(h_val % hash_size);

        if (!hash_table[tmp_hash_index]) {
            hash_table[tmp_hash_index] = new Cor_TupleStack();
            non_mt_bckts->push(tmp_hash_index);
        }

        hash_table[tmp_hash_index]->push(h_val,tuple);

#        ifdef DO_TRACE
            if (Cor_exEnv->dbg_indexing) {
                *(Cor_exEnv->trace_strm) << "HashIndex::insert_tuple: "
                    << "inserting in bucket "
                    << tmp_hash_index << endl;
            }
#        endif
    }

    // else (h_val == UnmatchedHashValue) { } 
    // i.e, pattern does not match iterator

    if (hash_count > limit)
        rehash_bigger();

    return(1);
}

// *************************************************************************
int Cor_HashIndex::delete_tuple(Cor_Tuple *tuple)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int del_tup = 0;
    HashVal h_val;

    h_val = hash(&(tuple->args()), NULL);

    if (h_val == VarHashValue) {

        del_tup |= var_chain->delete_tuple(tuple);

    } else {

        int tmp_hash_index = abs(h_val % hash_size);

        if (hash_table[tmp_hash_index]) {
            del_tup |= hash_table[tmp_hash_index]->delete_tuple(tuple);
            if (del_tup)
                hash_count--;
        }
    }

    return(del_tup);
}

// *************************************************************************
int  Cor_HashIndex::insert_entry(HashVal h_val, Cor_Tuple *tuple)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (tuple->is_deleted()) {
        // tuple has been logically deleted
        return(0);
    }

#    ifdef DO_TRACE
        if (Cor_exEnv->dbg_indexing) {
            *(Cor_exEnv->trace_strm) << "HashIndex::insert_entry: ";
            relation->name.print(NULL, Cor_exEnv->trace_strm);
            *(Cor_exEnv->trace_strm) << "(";
            tuple->printon(Cor_exEnv->trace_strm);
            *(Cor_exEnv->trace_strm) << ") ";

            if (argset != NULL) {
                // argument form index
                *(Cor_exEnv->trace_strm) << "[hash index on ";
                argset->printon(Cor_exEnv->trace_strm);
                *(Cor_exEnv->trace_strm) << " ]\n";
            } else {
                // pattern form index
                *(Cor_exEnv->trace_strm) << "[pattern matched ";
                pattern->print(NULL, Cor_exEnv->trace_strm);
                *(Cor_exEnv->trace_strm) << "] ";
                *(Cor_exEnv->trace_strm) << "[indexed on ";
                var_names->print(NULL, Cor_exEnv->trace_strm);
                *(Cor_exEnv->trace_strm) << " ]\n";
            }

            *(Cor_exEnv->trace_strm) << "        hash value is "
                                     <<  h_val << endl;
        }
#    endif

    if (h_val == VarHashValue) {
        // link it to the beginning of var_chain

        var_chain->push(h_val,tuple);

#        ifdef DO_TRACE
            if (Cor_exEnv->dbg_indexing) {
                *(Cor_exEnv->trace_strm) << "HashIndex::insert_tuple: ";
                *(Cor_exEnv->trace_strm) << "inserting in var_chain\n";
            }
#        endif
    }
    else if (h_val != UnmatchedHashValue) {
        // link it to the beginning of the appropriate hash bucket
        hash_count++;
        int tmp_hash_index = abs(h_val % hash_size);

        if (!hash_table[tmp_hash_index]) {
            hash_table[tmp_hash_index] = new Cor_TupleStack();
            non_mt_bckts->push(tmp_hash_index);
        }

        hash_table[tmp_hash_index]->push(h_val,tuple);

#        ifdef DO_TRACE
            if (Cor_exEnv->dbg_indexing) {
                *(Cor_exEnv->trace_strm) << "HashIndex::insert_entry: "
                    << "inserting in bucket "
                    << tmp_hash_index << endl;
            }
#        endif
    }

    // else (h_val == UnmatchedHashValue) { } 
    // i.e, pattern does not match iterator

    if (hash_count > limit)
        rehash_bigger();

    return(1);
}

// *************************************************************************
void Cor_HashIndex::rehash_bigger()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int i, j;
    Cor_TupleStack *bkt;

    //limit = (int) (limit * incr_ratio);
    int old_hash_size = hash_size;
    hash_size = (int)(hash_size * incr_ratio);

    non_mt_bckts->empty();

    // Recalculate limit to avoid roundoff errors
    limit = (int) (hash_size * max_occupancy);

    Cor_TupleStack **old_table = hash_table;
    hash_table = new Cor_TupleStack* [hash_size];

#ifdef    SPEED_HACK
    memset(hash_table, 0, sizeof(Cor_TupleStack*) * hash_size);
#else
    for (i=0; i < hash_size; i++)
        hash_table[i] = NULL;
#endif

    for (i=0; i < old_hash_size; i++) {
        if ((bkt=old_table[i])) {

            for (j=0; j < bkt->count(); j++) {
                if (bkt->tuples[j]) {
                    int tmp_hash_index = abs(bkt->hashs[j] % hash_size);

                    if (!hash_table[tmp_hash_index]) {
                        hash_table[tmp_hash_index] = new Cor_TupleStack();
                        non_mt_bckts->push(tmp_hash_index);
                    }

                    hash_table[tmp_hash_index]->push(bkt->hashs[j],
                                                    bkt->tuples[j]);
                }
            }

            delete old_table[i];
        }
    }

    delete [] old_table;
}

// *************************************************************************
//
//  iterator.ipos is used to keep track of where were are in the scan.
//  The possible values are:
//
//      #define Scanning_Init  0      Initial state 
//      #define Scanning_Known 1      scanning known bucket 
//      #define Scanning_Vars  2      scanning from var_chain 
//      #define Scanning_New   3      scanning from delta chain
//      #define Scanning_Done  4      done
//      #define Scanning_Hash  5      Scanning all buckets of hash table
//
// *************************************************************************

void Cor_HashIndex::initialize_iterator(Cor_TupleIterator& iterator)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
  iterator.stack_mark.get_mark();    

  // hash() retruns VarHashValue if iterator.arg_list.count == 0
  iterator.hash = hash(&iterator.arg_list, iterator.bindenv); 
  // changed to hashing on the appropriate set of arguments 

#    ifdef DO_TRACE
    if (Cor_exEnv->dbg_indexing) {
        *(Cor_exEnv->trace_strm)
            << "\nHashIndex::initialize_iterator: indexing ";
        iterator.relation->name.print(NULL, Cor_exEnv->trace_strm);
        *(Cor_exEnv->trace_strm) << " using ";
        iterator.arg_list.print(iterator.bindenv, Cor_exEnv->trace_strm);
    
        if (argset != NULL) { 
            // argument form index
            *(Cor_exEnv->trace_strm) << "[hash index on ";
            argset->printon(Cor_exEnv->trace_strm);
            *(Cor_exEnv->trace_strm) << " ]\n";
        } else { 
            // pattern form index
            *(Cor_exEnv->trace_strm) << "[pattern matched ";
            pattern->print(NULL, Cor_exEnv->trace_strm);
            *(Cor_exEnv->trace_strm) << "] ";
            *(Cor_exEnv->trace_strm) << "[indexed on ";
            var_names->print(NULL, Cor_exEnv->trace_strm);
            *(Cor_exEnv->trace_strm) << " ]\n";
        }
        *(Cor_exEnv->trace_strm)
            << "                 hash value is " << iterator.hash << endl;
    }
#    endif
    
    if (iterator.hash == UnmatchedHashValue) {
        iterator.ipos = Scanning_Done;

    } else if (iterator.hash != VarHashValue) {
        iterator.ipos = Scanning_Known;
        iterator.pindex = abs(iterator.hash % hash_size);
        iterator.bkt  = hash_table[iterator.pindex];
        iterator.ppos = 0;
    } else {

        // the entire relation has to be searched
        if (non_mt_bckts->is_empty()) {
            iterator.ipos = Scanning_Vars;
            iterator.bkt  = var_chain;
            iterator.ppos = 0;
            iterator.pindex = -1;
        } else {
            iterator.non_mt_indx = 0;
            iterator.pindex = non_mt_bckts->get_ith(iterator.non_mt_indx++);
            iterator.ipos = Scanning_Hash;
            iterator.bkt  = hash_table[iterator.pindex];
            iterator.ppos = 0;
        }
    }
}

// *************************************************************************
//  HashIndex::get_next_tuple which returns a Tuple* and 
//  HashIndex::get_next which returns a BindEnv* are extremely similar
//  in their code.  In fact, get_next_tuple duplicates the entire
//  code of get_next.  If trailing is used (i.e. DO_TRAIL is defined)
//  get_next modifies the iterator.bindenv AND returns a pointer to 
//  the BindEnv.  One way of making it consistent is to make it explicit
//  that iterator.bindenv will contain the modified BindEnv, and 
//  return a Tuple* consistently across the board.
// *************************************************************************
Cor_Tuple *Cor_HashIndex::get_next_tuple(Cor_TupleIterator& iterator)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif

    Cor_StackMark    stackmark0;
    Cor_Tuple       *result = NULL;

        // static to avoid repeated constructor calls
    static Cor_Term  left_term, right_term;
    Cor_CArg        *db_arg;
    HashVal          h_val;

    iterator.tuple_env = NULL;

    if (iterator.ipos == Scanning_Init) {
        initialize_iterator(iterator);
    } else {
        iterator.stack_mark.pop_to();
    }

    for (;;) {
#ifdef    SCAN_DEBUG
        cout << "looping through hash table, current state is: ";

        switch (iterator.ipos) {

        case Scanning_Known:
            cout << "Scanning_Known\n";
            break;

        case Scanning_Vars:
            cout << "Scanning_Vars\n";
            break;

        case Scanning_New:
            cout << "Scanning_New\n";
            break;

        case Scanning_Done: 
            cout << "Scanning_Done\n";
            break;
       
        case Scanning_Hash:
            cout << "Scanning_Hash\n";
            break;

        default:
            cout << "ERROR: Unknown state.\n";
            break;
        }
#endif

          // At the end of a hash bucket, what to do now.
        if (   (iterator.ipos == Scanning_Done)
            || !iterator.bkt || (iterator.ppos >= iterator.bkt->count() )) {

            switch (iterator.ipos) {

            case Scanning_Known:
                iterator.ipos = Scanning_Vars;
                iterator.ppos = 0;
                iterator.pindex = -1;
                iterator.bkt = var_chain;
                continue;
        
            case Scanning_Vars:
            case Scanning_New: // this should not occur
            case Scanning_Done: 
                return(NULL);
       
            default:
            case Scanning_Hash:
                // iterator.ipos == Scanning_Hash
                if (iterator.non_mt_indx < non_mt_bckts->count()) {
                    iterator.pindex =
                            non_mt_bckts->get_ith(iterator.non_mt_indx++);
                    iterator.bkt = hash_table[iterator.pindex];
                } else {
                    iterator.ipos = Scanning_Vars;
                    iterator.pindex = -1;
                    iterator.bkt = var_chain;
                }
                iterator.ppos = 0;
                continue;
            }
        }

        result = iterator.bkt->tuples[iterator.ppos];
        (iterator.ppos)++; 

        if (!result || result->is_deleted())
            continue;

        h_val = iterator.bkt->hashs[iterator.ppos-1];

        // Since many hash values could be mapped to the same hash bucket,
        // this preliminary check is more efficient.
        if ((iterator.ipos == Scanning_Known)
                    && (iterator.hash != h_val)) {
            continue;
        }

        // This section can call match, eq_chk, or assign

        int t_cnt = relation->arity();
        int l;

        for (l = 0; l < t_cnt; l++) {

#ifdef   UNIFY_VERBOSE 
            cout << "HI: Arglist is: " << iterator.arg_list << endl;
            cout << "Examining item: " << l << endl;
            cout << "Result tuple: " << *result << endl;

            cout << "iterator.bindenv: " << iterator.bindenv << "\n";
            iterator.bindenv->printon(&cout);
            cout << endl;
#endif

            Cor_Unify_Funct unifier;

            if (iterator.arg_unify)
                unifier = iterator.arg_unify[l];
            else
                unifier = COR_USE_UNIFY;

#ifdef    UNIFY_VERBOSE
            cout << "unifier is:" << iterator.arg_unify << endl;
#endif

            db_arg = &(result->args()[l]);

            int ans;

            switch (unifier) {

            case COR_USE_ASSIGN:
#ifdef    UNIFY_VERBOSE
                cout << "Assign.\n";

                cout << "it.arg_list[" << l << "]: "
                     << iterator.arg_list[l] << endl;
#endif
                ans = Cor_Assign_args(&(iterator.arg_list[l]),
                        iterator.bindenv, db_arg);

#ifdef    UNIFY_VERBOSE
                cout << "New iterator.bindenv: \n";
                iterator.bindenv->printon(&cout);
                cout << endl;
#endif
                break;

            case COR_USE_EQ_CHK:
#ifdef    UNIFY_VERBOSE
                cout << "Equal Check.\n";

                cout << "it.arg_list[" << l << "]: "
                     << iterator.arg_list[l] << endl;
#endif
                ans = Cor_EqlChk_args(&(iterator.arg_list[l]), 
                        iterator.bindenv, db_arg);
                break;

            case COR_USE_MATCH:
#ifdef    UNIFY_VERBOSE
                cout << "Match.\n";

                cout << "it.arg_list[" << l << "]: "
                     << iterator.arg_list[l] << endl;
#endif
                ans = Cor_Match_Arg( &(iterator.arg_list[l]),
                        iterator.bindenv, db_arg);
                break;

            case COR_USE_UNIFY:
            default: {
#ifdef    UNIFY_VERBOSE
                cout << "Unify.\n";

                cout << "it.arg_list[" << l << "]: "
                     << iterator.arg_list[l] << endl;
#endif
                    left_term.expr = iterator.arg_list[l];
                    left_term.bindenv = iterator.bindenv;
                    right_term.expr   = *db_arg;
                    right_term.bindenv = NULL;
                    ans = Cor_unify_args(left_term, right_term);
                    break;
                }
            }

            if (ans == COR_U_FAIL)
                break;
        }

        if (l == t_cnt) {
            iterator.reset_no_match();
            return(result);
        }

        stackmark0.pop_to();
    }

    return(NULL);
}

// *************************************************************************
Cor_BindEnv *Cor_HashIndex::get_next(Cor_TupleIterator& iterator)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_Tuple *tuple = get_next_tuple(iterator);

    if (tuple)    // Eliminate warning
        ;    // Nothing

    if (iterator.no_match())
        return(NULL);
    else
        return(iterator.bindenv);
}

// *************************************************************************
int Cor_HashIndex::is_subsumed_bkt(Cor_Tuple *newtup,
        HashVal hash_val, Cor_TupleStack *ts)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int k;
    if (ts) {

        for(k=0; k < ts->count(); k++) {
            if (ts->tuples[k]) {
                // Since many hash values could be mapped to
                // the same hash bucket, this preliminary
                // check is more efficient.
                if ((hash_val != VarHashValue)
                        && (hash_val != ts->hashs[k])
                        && (ts->hashs[k] != VarHashValue))
                    continue;

                // the tuple has been logically deleted
                if (ts->tuples[k]->is_deleted())
                    continue;

#ifdef    DEBUG
    //cout << "Checking subsumption of "    
    //     << *newtup << " vs " << *(ts->tuples[k]) << endl;
#endif

                if (Cor_subsumes_tuple(ts->tuples[k], newtup,
                        Cor_exEnv->dbg_indexing))
                    return(1);
            }
        }
    }

    return(0);
}

// *************************************************************************
int Cor_HashIndex::is_subsumed(Cor_Tuple *newtup)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    HashVal         hash_val;
    Cor_TupleStack *ts;
    int j, k, lim = non_mt_bckts->count();

    ASSERT(this);

    if (newtup->args().count() == 0) {
        // Normally this doesn't really mean that arity==0.
        // Rather, it means that the TupleIterator was
        // created with an ArgList parameter, which is a
        // However, in pathological cases, it might really
        // be a nullary Relation. Do the same thing in either case.
        hash_val = VarHashValue;
    } else {
        hash_val = (HashVal) hash(&(newtup->args()), NULL); 
    }

    if (hash_val == UnmatchedHashValue) {
        return(0);

    } else if (hash_val != VarHashValue) {

           // Only one bucket to look at.
        ts = hash_table[abs(hash_val % hash_size)];
        return( is_subsumed_bkt(newtup, hash_val, ts) );

    } else {

        // the entire relation has to be searched
        for (k=0; k < lim; k++) {

            j = non_mt_bckts->get_ith(k);

            if ( is_subsumed_bkt(newtup, hash_val, hash_table[j]) )
                return(1);
        }

        return( is_subsumed_bkt(newtup, hash_val, var_chain) );
    }

    return(0);
}

// *************************************************************************
