// *************************************************************************
//    HashIndexSet.C
// *************************************************************************

#include "HashIndexSet.h"

#include "Interface.h"
#include "Globals.h"
#include "HashSimpleRelation.h"
#include "Debug.h"

// *************************************************************************
//  This is an optimization to let a static environment be used in 
//  reentrant code.  We assume no parallelism however.

static Cor_BindEnv *static_env = NULL;
static int static_env_inuse=DONT_USE_STATIC_ENV;

// *************************************************************************
// Delete this tuple from all indices
int Cor_HashIndexSet::delete_tuple(Cor_Tuple *tuple)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int del_tup = 0;
    for (int i = 0; i < num_in_array; i++) {
        del_tup |= ((Cor_HashIndex *) Indices[i])->delete_tuple(tuple);
    }

    return(del_tup);
}

// *************************************************************************
//  WARNING::
//  Problem in HashIndexSet::find_exact
//  re-orderings of var_names is not taken into account while checking
//  whether two indexes are the same.  Thus, [V0|V1],V2 ; V1,V2 is
//  treated differently from [V0|V1],V2 ; V2,V1 though the two should
//  really be the same.  Merely sorting the var_names will not do the
//  trick since [V0|V1],V2 ; V1,V2 subsumes and is subsumed by
//  [V2|V1],V0 ; V1,V0 but sorting var_names would mess that up.
// *************************************************************************

int Cor_HashIndexSet::find_exact( Cor_HashIndex *index )
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    for (int i = 0; i < num_in_array; i++) {
        if (index->argset != NULL) {
            // argument form indexing
            if ( ((Cor_HashIndex *) Indices[i])->argset != NULL &&
                *index->argset == *((Cor_HashIndex *) Indices[i])->argset)
            {
                return COR_I_SUCCESS;
            }
        } else {
            // pattern form indexing
            if ( ((Cor_HashIndex *) Indices[i])->argset == NULL &&
                (index->num_vars_pat ==
                    ((Cor_HashIndex *)Indices[i])->num_vars_pat) &&
                (index->var_names->count() ==
                    ((Cor_HashIndex *) Indices[i])->var_names->count()))
            {
                Cor_StackMark stackmark;
                
                Cor_BindEnv pat_env1(index->num_vars_pat);
                Cor_BindEnv pat_env2(index->num_vars_pat);
                Cor_BindEnv pat_env3(index->num_vars_pat);
                Cor_BindEnv pat_env4(index->num_vars_pat);
                        
                // check if pattern and the var_names of each subsume
                // that of the other
                
                if (
                    (Cor_subsumes_arg_lists(
                        *concat_arglists(index->pattern,index->var_names),
                        *concat_arglists(((Cor_HashIndex *)
                                Indices[i])->pattern,
                        ((Cor_HashIndex *) Indices[i])->var_names),
                        &pat_env1,
                        &pat_env2) == COR_U_SUCCEED)
                    &&
                    (Cor_subsumes_arg_lists(
                        *concat_arglists(((Cor_HashIndex *)
                            Indices[i])->pattern,
                            ((Cor_HashIndex *) Indices[i])->var_names),
                        *concat_arglists(index->pattern, index->var_names),
                        &pat_env3,
                        &pat_env4) == COR_U_SUCCEED)
                    )
                {
                    stackmark.pop_to();
                    return COR_I_SUCCESS;
                } else {
                    stackmark.pop_to();
                }
            }
        }
    }
    return COR_I_FAIL;
}

// *************************************************************************
//  find an index that best matches the pattern contained arg_list, bind_env
//
//  WARNING:: The following code should at least check if the index is 
//  appropriate.  What it does is rather dangerous.  This is inappropriate
//  for programs with non-ground terms.  What is an appropriate index for one
//  invocation may not be appropriate for subsequent invocations : PRAVEEN
// *************************************************************************

Cor_GenericIndex *Cor_HashIndexSet::find_index(Cor_ArgList &arg_list,
    Cor_BindEnv *bindenv, int *tryindexp)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if ((tryindexp) && ((*tryindexp) >= 0) && ((*tryindexp) < num_in_array)) {
        // PRAVEEN: i checked that this gets entered, and yes it does !
        return ((Cor_GenericIndex *) Indices[*tryindexp]);
    }    

    if (!num_in_array) {
        *(Cor_exEnv->err_strm) << "There are no indices: Quitting\n";
        return NULL;
    }

    // find bitvector of pattern 
    int arity = arg_list.count();

    if (arity == 0) {
        // propositional case.
        if (rel->arity() != 0) {
            *(Cor_exEnv->err_strm) << 
                "Error: HashIndexSet::find_index: arity mismatch\n";
        }
        // Arbitrarily choose the 1st index and do a sequential scan.
        // need to ensure that the 1st index is an argument-type index.
        if (tryindexp)
            *tryindexp = 0;
        return (Cor_GenericIndex *) Indices[0];
    }

    Cor_BitVector adornment(arity);
    int num_bound = 0; // number of bound arg positions in pattern

    int i;
    for (i = 0; i < arity; i++) {

          // check which arguments are ground and set those bits to 1; rest 0
        if (arg_list[i].hash(bindenv) == VarHashValue) {
            adornment.set(i,0);
        } else {
            num_bound++;
            adornment.set(i,1);
        }
    }

    // Get the static BindEnv to use (if we don't have one yet).
    if (!static_env)
        static_env = new Cor_BindEnv(MAX_VARS);

    // Need to find "best" match 
    // For argument form indices: only those with FEWER ground arguments
    //     considered: i.e. index pattern subsumes pattern
    // For pattern form indices: index pattern should subsume pattern.
    // If all hash variables bound: give appropriate high weight
    // If only some bound: give a small weight say 0.5 because of
    //     greater selectivity than other trivial indices.

    Cor_GenericIndex *return_index = NULL;
    double max_weight = -1.0;

    for (i = 0; i < num_in_array; i++) {
        double cur_weight = 0.0;
        if (((Cor_HashIndex *) Indices[i])->argset != NULL) {
            // argument index
            if (arity != ((Cor_HashIndex *) Indices[i])->argset->len())
                continue;

            if ((num_bound == 0) &&
                !(((Cor_HashIndex *)Indices[i])->argset->any()))
            {
                // all free index
                if (tryindexp)
                    *tryindexp = i;
                return ((Cor_GenericIndex *)Indices[i]);                
            }

            if ((num_bound == arity) &&
                (adornment == *(((Cor_HashIndex *) Indices[i])->argset)))
            {
                // All arguments are bound - no better match can be found.
                if (tryindexp)
                    *tryindexp = i;
                return ((Cor_GenericIndex *)Indices[i]) ;
            }

            int j;
            for (j = 0; j < arity; j++) {

                if (adornment.test(j)) {
                    if (((Cor_HashIndex *) Indices[i])->argset->test(j)) {
                        cur_weight += 1.0;    // both are bound
                    }
                } else {
                    if (((Cor_HashIndex *) Indices[i])->argset->test(j)) {
                        // the index has more bound arguments; it is useless
                        cur_weight = 0.0; 
                        break;
                    }
                }
            }
        } else {
            // it is a pattern index
            Cor_StackMark stack_mark;
            Cor_BindEnv *patenv1;

#            define CLEANUP { \
                stack_mark.pop_to(); \
                if (--static_env_inuse != 0 || \
                    ((Cor_HashIndex *) Indices[i])->num_vars_pat > MAX_VARS) \
                { \
                    delete patenv1; \
                } \
            }

            if ((static_env_inuse++ == 0) &&
                ((Cor_HashIndex *) Indices[i])->num_vars_pat <= MAX_VARS)
            {
                patenv1 = static_env;
            } else {
                patenv1 = new Cor_BindEnv 
                    (((Cor_HashIndex *) Indices[i])->num_vars_pat);
            }

            if (Cor_subsumes_arg_lists(*((Cor_HashIndex *) Indices[i])->pattern,
                arg_list, patenv1, bindenv) == COR_U_SUCCEED)
            {
                int j;
                for (j = 0;
                       j < ((Cor_HashIndex *) Indices[i])->var_names->count(); 
                       j++)
                {
                    // The following is a heuristic for deciding what weight
                    // to assign to pattern form matches.  This will need to
                    // be updated as and when need be.

                  if ((*((Cor_HashIndex *)
                        Indices[i])->var_names)[j].hash(patenv1)
                                == VarHashValue)
                    {
                        cur_weight = 1.5;
                        CLEANUP
                        break;
                    } else {
                        cur_weight += 1.0;
                    }

                    if (cur_weight > arity) {
                        cur_weight = arity;
                        CLEANUP
                        break;
                    }
                }
            } else {
                // The pattern form index does not subsume pattern
                cur_weight = 0.0;
                CLEANUP
            }
        }

        if (cur_weight > max_weight) {
            // need to update information
            max_weight = cur_weight;
            if (tryindexp)
                *tryindexp = i;
            return_index = (Cor_GenericIndex *) Indices[i];
        }
    }
        
    if (!return_index) {
        // Have not been able to find any match
        // Arbitrarily choose the 1st index and do a sequential scan.
        // need to ensure that the 1st index is an argument-type index.
        if (tryindexp)
            *tryindexp = 0;
        return_index = (Cor_GenericIndex *) Indices[0];
    }

    return return_index;
}

// *************************************************************************
int Cor_HashIndexSet::insert_index( Cor_GenericIndex *index )
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif

    Cor_HashIndex *h_index = (Cor_HashIndex *) index;
    if (find_exact(h_index)) {
#        ifdef DO_TRACE
            if (Cor_exEnv->GlobalRelationOptions & REL_DISPLAY_INDEXOPS) {
                *(Cor_exEnv->trace_strm) << "Duplicate index: not inserted\n";
            }
#        endif
        return(COR_I_FAIL);
    }
    else {
        // new index to be inserted in list
        if (num_in_array == max_in_array) {
#            ifdef DEBUG
                *(Cor_exEnv->err_strm) << "Too many indices: not inserted\n";
#            endif
              return(COR_I_FAIL);
        }
        Indices[num_in_array++] = (Cor_GenericIndex *) h_index;
        // this is fine since array insertion starts from zero.
        return(COR_I_SUCCESS);
    }
}

// *************************************************************************
//  index is the index that will be used to fill up h_index. 
//  it should be an argument form index or a pattern form
//  index that indexes ALL the tuples in the relation, else
//  the filling up will not be correct.

void Cor_HashIndexSet::fill_index( Cor_GenericIndex *index )
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_HashIndex *h_index = (Cor_HashIndex *) index;
    Cor_HashIndex *f_index = (Cor_HashIndex *) Indices[0];

    // Don't want to fill up an index using itself
    ASSERT(h_index != f_index);

    // Now sequentially scan through f_index
    int j, k;
    Cor_TupleStack *bkt;

    for (j=0; j < f_index->hash_size; j++) {

        if ((bkt=f_index->hash_table[j])) {

            for (k=0; k < bkt->count(); k++) {
                if (bkt->tuples[k])
                    h_index->insert_tuple(bkt->tuples[k]);
            }
        }
    }

    if ((bkt=f_index->var_chain)) {
        for (k=0; k < bkt->count(); k++)
            if (bkt->tuples[k])
                h_index->insert_tuple(bkt->tuples[k]);
    }
}

// *************************************************************************
int Cor_HashIndexSet::add_all_tuples(Cor_StorageRelation *_reln,
    int marks_into_account, int check_dupl)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    // WARNING: duplicate checking is done only when marks are taken
    // into account.

    Cor_HashIndex *f_index = (Cor_HashIndex *) Indices[0];
    int dupl_check = check_dupl && _reln->check_subsum;

    if (!f_index->is_argument_form()) {
        *(Cor_exEnv->err_strm)
            << "ERROR: add_all_tuples does not have proper index.\n";
        return(COR_I_FAIL);
    }

    int j, k;
    Cor_TupleStack *bkt;

    if (!marks_into_account) {
        // Move nodes in the hash buckets

        int  ix, lim = f_index->non_mt_bckts->count();
        for (ix=0; ix < lim; ix++) {
            j = f_index->non_mt_bckts->get_ith(ix);

            if ((bkt=f_index->hash_table[j])) {
                for (k=0; k < bkt->count(); k++) {
                    if (bkt->tuples[k])
                        _reln->insert_raw(bkt->tuples[k]);
                }
            }
        }

        if ((bkt=f_index->var_chain)) {
            for (k=0; k < bkt->count(); k++) {
                if (bkt->tuples[k])
                    _reln->insert_raw(bkt->tuples[k]);
            }
        }
    } else {
        // same as above code except that insert_tuple has to be used

        int  ix, lim = f_index->non_mt_bckts->count();
        for (ix=0; ix < lim; ix++) {
            j = f_index->non_mt_bckts->get_ith(ix);

            if ((bkt=f_index->hash_table[j])) {
                for (k=0; k < bkt->count(); k++) {
                    if (bkt->tuples[k]) {
                        if (!dupl_check
                                  || !_reln->is_subsumed(bkt->tuples[k]))
                            _reln->insert_tuple(bkt->tuples[k]);
                    }
                }
            }
        }
      
        if ((bkt=f_index->var_chain)) {
            for (k=0; k < bkt->count(); k++) {
                if (bkt->tuples[k]) {
                    if (!dupl_check
                              || !_reln->is_subsumed(bkt->tuples[k]))
                        _reln->insert_tuple(bkt->tuples[k]);
                }
            }
        }
    }

    return(COR_I_SUCCESS);
}

// *************************************************************************
int Cor_HashIndexSet::add_all_goals(Cor_StorageRelation *_reln,
    Cor_Context *_context, Cor_StorageRelation *owner_rel)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    // This is used in Ordered Search
    // Tuples in (this) need to be converted to GoalNodes and added to
    // reln as well as context

    Cor_HashIndex *f_index = (Cor_HashIndex *) Indices[0];

    if (!f_index->is_argument_form()) {
        *(Cor_exEnv->err_strm)
            << "ERROR: add_all_goals does not have proper index.\n";
        return(COR_I_FAIL);
    }

    Cor_Tuple    *tuple = NULL;
    Cor_GoalNode *g_node = NULL;

    int j, k;
    Cor_TupleStack *bkt;

    for (j=0; j < f_index->hash_size; j++) {

        if ((bkt=f_index->hash_table[j])) {

            for (k=0; k < bkt->count(); k++) {
                if ((tuple=bkt->tuples[k])) {
                    if (!(tuple->is_deleted())) {
                        g_node = new Cor_GoalNode(tuple->_args,
                                    NULL, owner_rel);
                        g_node->set_no_args();
                        _reln->insert_tuple(g_node);
                        _context->add_new_goal(g_node, NULL);
                    }
                }
            }
        }
    }

    if ((bkt=f_index->var_chain)) {
        for (k=0; k < bkt->count(); k++)
            if ((tuple=bkt->tuples[k])) {
                if (!(tuple->is_deleted())) {
                    g_node = new Cor_GoalNode(tuple->_args, NULL, owner_rel);
                    g_node->set_no_args();
                    _reln->insert_tuple(g_node);
                    _context->add_new_goal(g_node, NULL);
                }
            }
    } 

    return(COR_I_SUCCESS);
}

// *************************************************************************
int Cor_HashIndexSet::move_all_entries(Cor_StorageRelation *_reln)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    //  WARNING: This code assumes that the two relations have exactly the
    //  same indices in exactly the same order, and both the relations
    //  are HashSimpleRelations

    Cor_HashSimpleRelation *otherrel = (Cor_HashSimpleRelation *) _reln;
    Cor_HashIndexSet *otherindset =  &(otherrel->index_set);
    
    if(num_in_array != otherindset->num_in_array) {
        *(Cor_exEnv->err_strm)
            << "CORAL:: internal error in move_all_entries\n";
        return -1;
    }

    for (int i=0; i < num_in_array; i++) {
        Indices[i]->move_all_entries(otherindset->Indices[i]);
    }
    return 1;
}

// *************************************************************************
