// *************************************************************************
//    ContextNode.C
// *************************************************************************

#include "Globals.h"
#include "ContextNode.h"
#include "StorageRelation.h"
#include "Debug.h"

// *************************************************************************
Cor_ContextNode::Cor_ContextNode(Cor_GoalNode *new_goal, Cor_Context *c) 
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    // assuming that a context node is created with only one unmarked goal

    goals = new_goal;
    all_next = NULL;
    all_prev = NULL;

    // set the ContextNode field of GoalNode 
    new_goal->c_node = this;

    // link in the new goal in the unmarked list
    Cor_GoalNodeLink *g_link = new Cor_GoalNodeLink(new_goal);
    first_unmarked_goal = g_link;
    last_unmarked_goal = g_link;

    first_marked_goal = NULL;
    last_marked_goal = NULL;

    mark = 0;                        // unmarked initially
    node_number = 0;

    negated = new_goal->negated;    // CHECK IF NEEDED??

    next = NULL;
    prev = NULL;
    next_marked = NULL;
    prev_marked = NULL;

    context = c;
}

// *************************************************************************
Cor_ContextNode::~Cor_ContextNode()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    free_all();
}

// *************************************************************************
int Cor_ContextNode::any_unmarked_goals()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    // NOTE: we also have to check for deleted subgoals, which will be 
    // retained in the goals if the context node is
    // non-singleton because deleting nodes from a set is difficult

    // this also tries and moves nodes from the unmarked goals list to
    // the marked goals list so that space can be reclaimed later
    // NOTE: should not affect any asymptotics

    if (first_unmarked_goal == NULL) return 0;        // failure

    for (Cor_GoalNodeLink *g_ptr = first_unmarked_goal; g_ptr != NULL; ) {
        if (!(g_ptr->g_node->is_deleted()))
            return(1);        // success        
        else {
            // the node pointed to by g_ptr is deleted
            // deallocate the node and update first_unmarked_goal
            first_unmarked_goal = g_ptr->next;
            move_to_marked_list(g_ptr);
            // NOTE: should not do a "delete g_ptr"
            // because that would result in unreclaimable space
        }
    }
    update_goalnode_info();
    return(0);
}

// *************************************************************************
Cor_GoalNode *Cor_ContextNode::get_unmarked_goal()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (!any_unmarked_goals()) { 
        return NULL;
    }

    // choose the first goal node from unmarked list, mark it, 
    // update unmarked and marked lists
    Cor_GoalNodeLink *g_ptr = first_unmarked_goal;
    first_unmarked_goal = first_unmarked_goal->next;
    move_to_marked_list(g_ptr);
    update_goalnode_info();

    Cor_GoalNode *g = g_ptr->g_node;
    g->mark_goal();
        // marking context node and updating Context done from caller site
        // as is updating the node_number field of this node

    return g;
}

// *************************************************************************
void Cor_ContextNode::move_goals_to_done()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    ASSERT((is_marked()));
    ASSERT((first_unmarked_goal == NULL));

    // potentially there are two types of goals in the marked list
    // (1) magic facts: these have to be moved to the done relation
    // (2) supplementary facts: probably nothing has to be done to these

    for (Cor_GoalNodeLink *g_tmp = first_marked_goal; g_tmp != NULL;
        g_tmp = g_tmp->next)
    {
        if ((g_tmp->g_node->is_magic_goal())&&(!(g_tmp->g_node->is_deleted())))
        {
            Cor_StorageRelation *mp_done = g_tmp->g_node->find_done_relation();
#            ifdef DO_TRACE
                if (Cor_exEnv->dbg_ordsearch) {
                    *(Cor_exEnv->trace_strm) << 
                        "[Inserting " <<  mp_done->name.symbol() << "(";
                    g_tmp->g_node->args().print(NULL,Cor_exEnv->trace_strm);
                    context->print_all(Cor_exEnv->trace_strm);
                }
#            endif

            // Make a copy of the tuple, but not its args.
            // Inform the tuple that it doesn't own its args 
            // (so they are not deleted later)
            Cor_Tuple *mytup = new Cor_Tuple(g_tmp->g_node->_args);
            mytup->set_no_args();

            mp_done->insert_tuple(mytup);
            g_tmp->g_node->owner_rel = mp_done;
        }
    }
}

// *************************************************************************
void Cor_ContextNode::move_to_marked_list(Cor_GoalNodeLink *g_ptr)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (first_marked_goal == NULL) {
        // create a new list
        first_marked_goal = g_ptr;
        last_marked_goal = g_ptr;
    }
    else {
        // insert at end of list
        last_marked_goal->next = g_ptr;
        last_marked_goal = g_ptr;
    }
    last_marked_goal->next = NULL;
}

// *************************************************************************
void Cor_ContextNode::merge_nodes(Cor_ContextNode *tbm)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    // the fields of tbm have to be merged into "this"

    // merge goals
    goals = goals->link_goals(tbm->goals);

    // merge unmarked goal lists
    if (first_unmarked_goal == NULL) {
        // nothing unmarked in "this"
        first_unmarked_goal = tbm->first_unmarked_goal;
        last_unmarked_goal = tbm->last_unmarked_goal;
    }
    else {
        if (tbm->first_unmarked_goal != NULL) {
            // something unmarked in tbm
            last_unmarked_goal->next = tbm->first_unmarked_goal;
            last_unmarked_goal = tbm->last_unmarked_goal;
        }
    }

    // merge marked goal lists
    ASSERT((first_marked_goal != NULL)); 
    if (tbm->first_marked_goal != NULL) {
        // something marked in tbm
        last_marked_goal->next = tbm->first_marked_goal;
        last_marked_goal = tbm->last_marked_goal;
    }

    // update node_number of the merged node to the higher of the two
    ASSERT((tbm->node_number > node_number));
    node_number = tbm->node_number;

    // NOTE: make sure caller in Context updates the first, last, etc. fields

    // NOTE: freeing up space of merged in node should be done from caller
}

// *************************************************************************
void Cor_ContextNode::update_goalnode_info()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (first_unmarked_goal == NULL)
        last_unmarked_goal = NULL;
    if (first_marked_goal == NULL)
        last_marked_goal = NULL;
}

// *************************************************************************
void Cor_ContextNode::free_all()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    // frees up the space of GoalNodeLink's but not the GoalNode's 
    // themselves because they are in a separate index and freeing
    // them would result in dangling pointers

    goals = NULL;

    Cor_GoalNodeLink *cur, *prv = NULL;
    for (cur = first_unmarked_goal; cur; cur=cur->next ) {
        delete prv;
        prv = cur;
    }
    first_unmarked_goal = NULL;
    last_unmarked_goal = NULL;

    for (cur = first_marked_goal; cur; cur=cur->next ) {
        delete prv;
        prv = cur;
    }
    delete prv;
    first_marked_goal = NULL;
    last_marked_goal = NULL;

    mark = 0;
    node_number = 0;

    next = NULL;
    prev = NULL;
    next_marked = NULL;
    prev_marked = NULL;

    context = NULL;
}

// *************************************************************************
void Cor_ContextNode::print_all(ostream *file)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    *file << "\tPrinting ContextNode:\n";
    *file << "\t\tMark = " << mark << ";  Node number = " << node_number 
          << ".\n";

    Cor_GoalNodeLink *temp;

    if (first_unmarked_goal == NULL) {
        *file << "\t\tNo unmarked goals.\n";
    } else {
        // there are unmarked goals
        *file << "\t\tList of unmarked goals =\n";
        for (temp = first_unmarked_goal; temp != last_unmarked_goal->next;
            temp = temp->next)
        {
            temp->g_node->print_all(file);
        }
    }

    if (first_marked_goal == NULL) {
        *file << "\t\tNo marked goals.\n";
    }
    else { // there are marked goals
        *file << "\t\tList of marked goals =\n";
        for (temp = first_marked_goal; temp != last_marked_goal->next;
                                        temp = temp->next) {
            temp->g_node->print_all(file);
        }
    }
    *file << endl;
}

// *************************************************************************
