// *************************************************************************
//    Operators.C
// *************************************************************************

#include <math.h>

#include "Operators.h"

#include "Globals.h"
#include "Interface.h"
#include "Tuple.h"
#include "TupleIterator.h"
#include "BindEnv.h"
#include "CoralError.h"
#include "Debug.h"


// *************************************************************************
Cor_Tuple *Cor_FuncOp(Cor_Tuple *query)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (query->arity() != 3) {
        *(Cor_exEnv->err_strm) << 
                "Three arguments required for Functor.\n";
        return NULL;
    }

    Cor_CArg arg0 = (query->args())[0];
    Cor_CArg arg1 = (query->args())[1];
    Cor_CArg arg2 = (query->args())[2];

    if ( (! arg0.is_var()) && (! arg0.is_functor())
        && (! arg0.is_symbol()) ) {
        *(Cor_exEnv->err_strm) << 
                "First argument should be a variable or functor\n";
        return NULL;
    }
    else if ((arg0.is_functor())
            || (arg0.is_symbol())) {

        Cor_CArg func_name;
        Cor_CArg num;

        if (arg0.is_symbol()) {
            func_name = arg0;
            num = 0;
        } else {
            func_name = Cor_CArg(arg0.functor());
            Cor_ArgList *func_list = functor_args(&arg0);
            num = func_list->arity();
        }

        Cor_ArgList *alist = new Cor_ArgList(3);
        (*alist)[0] = arg0;
        (*alist)[1] = func_name;
        (*alist)[2] = num;

        Cor_Tuple *result = new Cor_Tuple(alist);
        return result;
    }
    else if (arg0.is_var() && arg1.is_symbol() && arg2.is_num()) {
        if (! arg2.is_int()) {
            *(Cor_exEnv->err_strm) << 
                "Third argument should be of integer type.\n";
            return NULL;
        }

        int index = arg2.NumIval();
        Cor_CArg *new_arg = new Cor_CArg(arg1, index);
        int i;
        for (i = 0; i < index; i++)
            (new_arg->args())[i] = i;
   
        Cor_ArgList *alist = new Cor_ArgList(3);
        (*alist)[0] = *new_arg;
        (*alist)[1] = arg1;
        (*alist)[2] = arg2;

        Cor_Tuple *result = new Cor_Tuple(alist);
        return(result);
    }

    *(Cor_exEnv->err_strm) << "Wrong argument types supplied to Functor\n";
    return NULL;
}

// *************************************************************************
Cor_Tuple *Cor_ArgOp(Cor_Tuple *query)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (query->arity() != 3) {
        *(Cor_exEnv->err_strm) << "Three arguments required for Arg.\n";
        return(NULL);
    }

    Cor_CArg arg0 = (query->args())[0];
    Cor_CArg arg1 = (query->args())[1];
    Cor_CArg arg2 = (query->args())[2];

    if (arg0.is_num() && arg1.is_functor() ) {
        if (!arg0.is_int()) {
                *(Cor_exEnv->err_strm) << 
                    "First argument should be of integer type.\n";
                return(NULL);
        }
        Cor_ArgList *func_list = functor_args(&arg1);
        int num = make_int(arg0);
        if (num > func_list->arity()) {
            *(Cor_exEnv->err_strm) << "Not enough arguments in arg.\n";
            return(NULL);
        }
        Cor_CArg new_arg = func_list->arg(num-1);
        Cor_ArgList *alist = make_arglist(arg0, arg1, new_arg);

        Cor_Tuple *result = new Cor_Tuple(alist);
        return(result);
    }

    *(Cor_exEnv->err_strm) << "Wrong argument types supplied to arg().\n";
    return(NULL);
}

// *************************************************************************
void Cor_AddOp(Cor_CArg& carg, const Cor_CArg& arg0, const Cor_CArg& arg1)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if ( !arg0.is_num() || !arg1.is_num() ) {
        *(Cor_exEnv->err_strm) << 
            "One of the arguments to AddOp is not a number.\n";
        *(Cor_exEnv->err_strm) << "Args are: " << arg0 << " and "
                               << arg1 << endl;
        carg = Cor_NullCArg;
        return;
    }

    if (arg0.is_int() && arg1.is_int()) {
        carg =arg0.NumIval() + arg1.NumIval();
        return;
    }

    carg = arg0.NumDval() + arg1.NumDval();
}

// *************************************************************************
void Cor_SubOp(Cor_CArg& carg, const Cor_CArg& arg0, const Cor_CArg& arg1)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if ( !arg0.is_num() || !arg1.is_num() ) {
        *(Cor_exEnv->err_strm) << 
            "One of the arguments to SubOp is not a number.\n";
        carg = Cor_NullCArg;
        return;
    }

    if (arg0.is_int() && arg1.is_int())
      carg = arg0.NumIval() - arg1.NumIval();
    else
      carg = arg0.NumDval() - arg1.NumDval();
}

// *************************************************************************
void  Cor_MulOp(Cor_CArg& carg, const Cor_CArg& arg0,const Cor_CArg& arg1)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if ( !arg0.is_num() || !arg1.is_num() ) {
        *(Cor_exEnv->err_strm) << 
            "One of the arguments to MulOp is not a number.\n";
        carg = Cor_NullCArg;
        return;
    }

    if (arg0.is_int() && arg1.is_int()) {
        carg = arg0.NumIval() *arg1.NumIval();
        return;
    }

    carg = arg0.NumDval() *arg1.NumDval();
}

// *************************************************************************
void Cor_DivOp(Cor_CArg& carg, const Cor_CArg& arg0,const Cor_CArg& arg1)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if ( !arg0.is_num() || !arg1.is_num() ) {
        *(Cor_exEnv->err_strm) << 
            "One of the arguments to DivOp is not a number.\n";
        carg = Cor_NullCArg;
        return;
    }

    if (!arg1.compare_num(Cor_CArg(0))) {
        *(Cor_exEnv->err_strm) << 
            "Divided by Zero!\n";
        carg = Cor_NullCArg;
        return;
    }

    if (arg0.is_int() && arg1.is_int()) {
        carg =arg0.NumIval() / arg1.NumIval();
        return;
    }

    carg =  arg0.NumDval() / arg1.NumDval();
}

// *************************************************************************
void Cor_SumOp(Cor_CArg& carg, Cor_Relation *rel)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int total_i = 0;
    double total_d = 0.0;
    int isInt = 0;
    int isDoub = 0;

    carg = Cor_NullCArg;

    Cor_StackMark stack_mark0;
    Cor_BindEnv *env0 = new Cor_BindEnv(rel->arity());
    Cor_ArgList *args = make_vararglist(rel->arity());
    Cor_TupleIterator *iter = new Cor_TupleIterator(rel, *args, env0);
    for(;;) {
        Cor_Tuple *ret_tuple = iter->get_next_tuple();
        
        if (iter->no_match())
            break;
        
        if ((ret_tuple->arity() != 1)
            || ( ! ret_tuple->arg(0).is_num() )) {
            *(Cor_exEnv->err_strm) << 
              "Error: Sum: Set must contain numeric constants as elements.\n";
            carg = *Cor_ErrorSymbol;
            return;
        }
        
        if (ret_tuple->arg(0).is_int()) {
            isInt = 1;
            total_i += ret_tuple->arg(0).NumIval();
        }
        else if (ret_tuple->arg(0).is_double()) {
            isDoub = 1;
            total_d += ret_tuple->arg(0).NumDval();
        }
        else {
            *(Cor_exEnv->err_strm) << 
              "Error: Sum: Set must contain numeric constants as elements.\n";
            carg = *Cor_ErrorSymbol;
            return;
        }
    }

    stack_mark0.pop_to();

    if (isInt && isDoub) {
        *(Cor_exEnv->err_strm) << 
            "Error: Sum: Set must contain either integers or doubles\n";
        carg = *Cor_ErrorSymbol;
        return;
    }

    if (isInt)
        carg = total_i;
    if (isDoub)
        carg = total_d;
}

// *************************************************************************
void Cor_ProdOp(Cor_CArg& carg, Cor_Relation *rel)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int total_i = 1;
    double total_d = 1.0;
    int is_int = 0;
    int is_doub = 0;

    carg = Cor_NullCArg;

    Cor_StackMark stack_mark0;
    Cor_BindEnv *env0 = new Cor_BindEnv(rel->arity());
    Cor_ArgList *args = make_vararglist(rel->arity());
    Cor_TupleIterator *iter = new Cor_TupleIterator(rel, *args, env0);
    for(;;) {
        Cor_Tuple *ret_tuple = iter->get_next_tuple();
        
        if (iter->no_match())
            break;
        
        if ((ret_tuple->arity() != 1)
            || ( ! ret_tuple->arg(0).is_num() )) {
            *(Cor_exEnv->err_strm) << 
              "Error: Prod: Set must contain numeric constants as elements.\n";
            carg = *Cor_ErrorSymbol;
            return;
        }
        
        if (ret_tuple->arg(0).is_int()) {
            is_int = 1;
            total_i *= ret_tuple->arg(0).NumIval();
        }
        else if (ret_tuple->arg(0).is_double()) {
            is_doub = 1;
            total_d *= ret_tuple->arg(0).NumDval();
        }
        else {
            *(Cor_exEnv->err_strm) << 
              "Error: Sum: Set must contain numeric constants as elements.\n";
            carg = *Cor_ErrorSymbol;
            return;
        }
    }

    stack_mark0.pop_to();

    if (is_int && is_doub) {
        *(Cor_exEnv->err_strm) << 
            "Error: Sum: Set must contain either integers or doubles\n";
        carg = *Cor_ErrorSymbol;
        return;
    }

    if (is_int)
        carg = total_i;
    if (is_doub)
        carg = total_d;
}

// *************************************************************************
void Cor_AvgOp(Cor_CArg& carg, Cor_Relation *rel)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    double total_d = 0.0, avg_d=0.0;
    int count = 0;

    Cor_StackMark stack_mark0;
    Cor_BindEnv *env0 = new Cor_BindEnv(rel->arity());
    Cor_ArgList *args = make_vararglist(rel->arity());
    Cor_TupleIterator *iter = new Cor_TupleIterator(rel, *args, env0);
    for(;;) {
        Cor_Tuple *ret_tuple = iter->get_next_tuple();
        
        if (iter->no_match())
            break;
        
        if ((ret_tuple->arity() != 1)
            || ( ! ret_tuple->arg(0).is_num() )) {
            *(Cor_exEnv->err_strm) << 
              "Error: Sum: Set must contain numeric constants as elements.\n";
            carg = *Cor_ErrorSymbol;
            return;
        }
        
        count++;
        if (ret_tuple->arg(0).is_int()) {
            total_d += (double) ret_tuple->arg(0).NumIval();
        }
        else if (ret_tuple->arg(0).is_double()) {
            total_d += ret_tuple->arg(0).NumDval();
        } else {
            *(Cor_exEnv->err_strm) << 
              "Error: Sum: Set must contain numeric constants as elements.\n";
            carg = *Cor_ErrorSymbol;
            return;
        }
    }

    stack_mark0.pop_to();

    if (count) 
        avg_d = total_d / (double) count;
    carg = avg_d;
}

// *************************************************************************
void Cor_StdDevOp(Cor_CArg& carg, Cor_Relation *rel)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    double sum = 0.0, avg, stddev;
    double sum_of_sqrs = 0.0, val;
    int count = 0;

    Cor_StackMark stack_mark0;
    Cor_BindEnv *env0 = new Cor_BindEnv(rel->arity());
    Cor_ArgList *args = make_vararglist(rel->arity());
    Cor_TupleIterator *iter = new Cor_TupleIterator(rel, *args, env0);

    for(;;) {
        Cor_Tuple *ret_tuple = iter->get_next_tuple();
        
        if (iter->no_match())
            break;
        
        if ((ret_tuple->arity() != 1)
            || ( ! ret_tuple->arg(0).is_num() )) {
            *(Cor_exEnv->err_strm) << 
              "Error: Sum: Set must contain numeric constants as elements.\n";
            carg = *Cor_ErrorSymbol;
            return;
        }
        
        count++;
        if (ret_tuple->arg(0).is_int()) {
            val = (double) ret_tuple->arg(0).NumIval();
        } else if (ret_tuple->arg(0).is_double()) {
            val = ret_tuple->arg(0).NumDval();
        } else {
            *(Cor_exEnv->err_strm)
             << "Error: Sum: Set must contain numeric constants as elements.\n";
            carg = *Cor_ErrorSymbol;
            return;
        }

        sum += val;
        sum_of_sqrs += (val * val);
    }

    stack_mark0.pop_to();

    if (count > 1) {
        avg = sum / (double) count;
        double decr = (double) (count-1);

        double tmp = (sum_of_sqrs / decr)
                   - ((((double) count) * avg * avg) / decr);
        stddev = sqrt(tmp);

    } else 
        stddev = 0.0;

    carg = stddev;
}

// *************************************************************************
void Cor_CountOp(Cor_CArg& carg, Cor_Relation *rel)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    carg = rel->count;
    return;
}

// *************************************************************************
void Cor_MinOp(Cor_CArg& carg, Cor_Relation *rel)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int first_tuple_flag = 1;
    Cor_Tuple *ret_tuple;
    Cor_CArg min;
    Cor_StackMark stack_mark0;
    Cor_BindEnv *env0 = new Cor_BindEnv(rel->arity());
    Cor_ArgList *args = make_vararglist(rel->arity());
    Cor_TupleIterator *iter = new Cor_TupleIterator(rel, *args, env0);

    for(;;) {
        if (first_tuple_flag) {
            ret_tuple = iter->get_next_tuple();
            if (iter->no_match()) {
                *(Cor_exEnv->err_strm) << 
                    "Error: Min: No elements in the set.\n";
                carg = Cor_NullCArg; return;
            }
            min = ret_tuple->arg(0);
            first_tuple_flag = 0;
        }
        else
            ret_tuple = iter->get_next_tuple();

        if (iter->no_match())
            break;

        if ((ret_tuple->arity() != 1)
            || ( ! ret_tuple->arg(0).is_num() )) {
            *(Cor_exEnv->err_strm) << 
              "Error: Min: Set must contain numeric constants as elements.\n";
            carg = *Cor_ErrorSymbol;
            return;
        }

        if (!first_tuple_flag) {
            Cor_Term *num0 = new Cor_Term(min) ;
            Cor_Term *num1 = new Cor_Term(ret_tuple->arg(0));
            if (!Cor_LssOp(*num0, *num1))
                min = ret_tuple->arg(0);
        }
    }
    stack_mark0.pop_to();
    carg = min;
}

// *************************************************************************
void Cor_MaxOp(Cor_CArg& carg, Cor_Relation *rel)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int first_tuple_flag = 1;
    Cor_Tuple *ret_tuple;
    Cor_CArg max;
    Cor_StackMark stack_mark0;
    Cor_BindEnv *env0 = new Cor_BindEnv(rel->arity());
    Cor_ArgList *args = make_vararglist(rel->arity());
    Cor_TupleIterator *iter = new Cor_TupleIterator(rel, *args, env0);

    for(;;) {
        if (first_tuple_flag) {
            ret_tuple = iter->get_next_tuple();
            if (iter->no_match()) {
                *(Cor_exEnv->err_strm) << 
                    "Error: Max: No elements in the set.\n";
                carg = Cor_NullCArg; return;
            }
            max = ret_tuple->arg(0);
            first_tuple_flag = 0;
        }
        else
            ret_tuple = iter->get_next_tuple();

        if (iter->no_match())
            break;

        if ( (ret_tuple->arity() != 1)
                || ( ! ret_tuple->arg(0).is_num() )) {
            *(Cor_exEnv->err_strm) << 
              "Error: Max: Set must contain numeric constants as elements.\n";
            carg = *Cor_ErrorSymbol;
            return;
        }

        if (!first_tuple_flag) {
            Cor_Term *num0 = new Cor_Term(max) ;
            Cor_Term *num1 = new Cor_Term(ret_tuple->arg(0));
            if (!Cor_GrtOp(*num0, *num1))
                max = ret_tuple->arg(0);
        }
    }
    stack_mark0.pop_to();
    carg =  max;
}

// *************************************************************************
int Cor_EquOp(Cor_Term& term0, Cor_Term& term1)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    return Cor_unify_args(term0, term1);
}

// *************************************************************************
int Cor_NeqOp(Cor_Term& term0, Cor_Term& term1)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    return !Cor_unify_args(term0, term1);
}

// *************************************************************************
int Cor_LssOp(Cor_Term& term0, Cor_Term& term1)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    return (term0.expr.compare_num(term1.expr) < 0);
}

// *************************************************************************
int Cor_GrtOp(Cor_Term& term0, Cor_Term& term1)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    return (term0.expr.compare_num(term1.expr) > 0);
}

// *************************************************************************
int Cor_LeqOp(Cor_Term& term0, Cor_Term& term1)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    return (term0.expr.compare_num(term1.expr) <= 0);
}

// *************************************************************************
int Cor_GeqOp(Cor_Term& term0, Cor_Term& term1)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    return (term0.expr.compare_num(term1.expr) >= 0);
}

// *************************************************************************
void Cor_LShiftOp(Cor_CArg& carg, const Cor_CArg& arg0, const Cor_CArg& arg1)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif

    if (arg0.is_int() && arg1.is_int())
        carg = arg0.NumIval() << arg1.NumIval();
    else{
        *(Cor_exEnv->err_strm) << 
            "One of the arguments to LShiftOp is not an integer.\n";
        carg = Cor_NullCArg; 
    }
}

// *************************************************************************
void Cor_RShiftOp(Cor_CArg& carg, const Cor_CArg& arg0, const Cor_CArg& arg1)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (arg0.is_int() && arg1.is_int())
        carg = arg0.NumIval() >> arg1.NumIval();
    else{
       *(Cor_exEnv->err_strm) << 
           "One of the arguments to RShiftOp is not an integer.\n";
       carg = Cor_NullCArg; 
    }
}

// *************************************************************************
void Cor_BitorOp(Cor_CArg& carg, const Cor_CArg& arg0, const Cor_CArg& arg1)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (arg0.is_int() && arg1.is_int())
        carg =  arg0.NumIval() | arg1.NumIval();
    else{
        *(Cor_exEnv->err_strm) << 
            "One of the arguments to BitorOp is not an integer.\n";
        carg = Cor_NullCArg; 
    }
}

// *************************************************************************
void Cor_BitandOp(Cor_CArg& carg, const Cor_CArg& arg0, const Cor_CArg& arg1)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (arg0.is_int() && arg1.is_int())
        carg =  arg0.NumIval() & arg1.NumIval();
    else{
        *(Cor_exEnv->err_strm) << 
            "One of the arguments to BitandOp is not an integer.\n";
        carg = Cor_NullCArg; 
    }
}

// *************************************************************************
void Cor_BitxorOp(Cor_CArg& carg, const Cor_CArg& arg0, const Cor_CArg& arg1)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (arg0.is_int() && arg1.is_int())
        carg =arg0.NumIval() ^ arg1.NumIval();
    else{
        *(Cor_exEnv->err_strm) << 
            "One of the arguments to BitxorOp is not an integer.\n";
        carg = Cor_NullCArg; 
    }
}

// *************************************************************************
void Cor_ModOp(Cor_CArg& carg, const Cor_CArg& arg0, const Cor_CArg& arg1)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (arg0.is_int() && arg1.is_int())
        carg =  arg0.NumIval() % arg1.NumIval();
    else{
        *(Cor_exEnv->err_strm) << 
            "One of the arguments to ModOp is not an integer.\n";
        carg = Cor_NullCArg; 
    }
}

// *************************************************************************
