// *************************************************************************
// SocketRelSolver.C
// solver function for builtin: socket_rel
// *************************************************************************

// user_command: socket_rel
// syntax: socket_rel(local_name, arity (or func-arg), host, port, work_space, relation name)
//     or socket_rel(local_name(_,_....), host, port, work_space, relation name)
// description: socket_rel defines a local relation which is stored in
//     a remote host, accessed via a socket port, in a work_space, with a
//     remote name.
//error_handling: the following situation will cause an error msg echoed 
//     to the user, not when 
//      1) unknown host, unknown port, connection lost
//      2) wrong workspace, relation name, argument number 


#include "Globals.h"
#include "CoralError.h"
#include "BuiltinRelation.h"
#include "TupleIterator.h"
#include "SockRelation.h"

#define FAIL_RETURN  {iterator.set_no_match(); return(NULL);}

// *************************************************************************
Cor_BindEnv *SocketRelSolver(Cor_BuiltinRelation&,
        Cor_TupleIterator& iterator)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
#ifdef SOCK_REL

    char *local_name, *host, *relation_name, *work_space;
    int arity, port;
    
    if (iterator.arg_list.count() != 6 && iterator.arg_list.count() !=5)  {
        CORAL_error(COR_BAD_ARITY, "5 or 6 arguments to socket_rel needed", "SocketRelSolver");
        FAIL_RETURN;
    }
    
    if (iterator.arg_list.count() == 6) {
        Cor_Term term1 (iterator.arg_list[0], iterator.bindenv);
        Cor_deep_dereference_term (term1);
        Cor_Term term2 (iterator.arg_list[1], iterator.bindenv);
        Cor_deep_dereference_term (term2);
        Cor_Term term3 (iterator.arg_list[2], iterator.bindenv);
        Cor_deep_dereference_term (term3);
        Cor_Term term4 (iterator.arg_list[3], iterator.bindenv);
        Cor_deep_dereference_term (term4);
        Cor_Term term5 (iterator.arg_list[4], iterator.bindenv);
        Cor_deep_dereference_term (term5);
        Cor_Term term6 (iterator.arg_list[5], iterator.bindenv);
        Cor_deep_dereference_term (term6);

        if (!term1.expr.is_symbol()) {
            CORAL_error(COR_BAD_ARGLIST, "Local name must be a symbol","SocketRelSolver");
            term1.printon(Cor_exEnv->err_strm);
            *(Cor_exEnv->err_strm) << endl;
            FAIL_RETURN;
        }

        if (term2.expr.is_functor()) 
            arity = term2.expr.arity();
        else if (term2.expr.is_num())
            arity = term2.expr.NumIval();
        else {
            CORAL_error(COR_BAD_ARGLIST, "Arity must be a number or func-arg","SocketRelSolver");
            term2.printon(Cor_exEnv->err_strm);
            *(Cor_exEnv->err_strm) << endl;
            FAIL_RETURN;
        }
        
        if (!term3.expr.is_symbol()) {
            CORAL_error(COR_BAD_ARGLIST, "Host must be a symbol","SocketRelSolver");
            term3.printon(Cor_exEnv->err_strm);
            *(Cor_exEnv->err_strm) << endl;
            FAIL_RETURN;
        }
        if (!term4.expr.is_num()) {
            CORAL_error(COR_BAD_ARGLIST, "Port must be a number","SocketRelSolver");
            term4.printon(Cor_exEnv->err_strm);
            *(Cor_exEnv->err_strm) << endl;
            FAIL_RETURN;
        }
        if (!term5.expr.is_symbol()) {
            CORAL_error(COR_BAD_ARGLIST, "Workspace must be a symbol","SocketRelSolver");
            term5.printon(Cor_exEnv->err_strm);
            *(Cor_exEnv->err_strm) << endl;
            FAIL_RETURN;
        }
        if (!term6.expr.is_symbol()) {
            CORAL_error(COR_BAD_ARGLIST, "Relation name must be a symbol","SocketRelSolver");
            term6.printon(Cor_exEnv->err_strm);
            *(Cor_exEnv->err_strm) << endl;
            FAIL_RETURN;
        }
       
        local_name    = term1.expr.symbol();
        host          = term3.expr.symbol();
        port          = term4.expr.NumIval();
        work_space    = term5.expr.symbol();
        relation_name = term6.expr.symbol();

    }
    else {
        Cor_Term term1 (iterator.arg_list[0], iterator.bindenv);
        Cor_deep_dereference_term (term1);
        Cor_Term term2 (iterator.arg_list[1], iterator.bindenv);
        Cor_deep_dereference_term (term2);
        Cor_Term term3 (iterator.arg_list[2], iterator.bindenv);
        Cor_deep_dereference_term (term3);
        Cor_Term term4 (iterator.arg_list[3], iterator.bindenv);
        Cor_deep_dereference_term (term4);
        Cor_Term term5 (iterator.arg_list[4], iterator.bindenv);
        Cor_deep_dereference_term (term5);
        
        if (!term1.expr.is_functor()) {
            CORAL_error(COR_BAD_ARGLIST, "Local name must be a functor","SocketRelSolver");
            term1.printon(Cor_exEnv->err_strm);
            *(Cor_exEnv->err_strm) << endl;
            FAIL_RETURN;
        }

        if (!term2.expr.is_symbol()) {
            CORAL_error(COR_BAD_ARGLIST, "Host must be a symbol","SocketRelSolver");
            term2.printon(Cor_exEnv->err_strm);
            *(Cor_exEnv->err_strm) << endl;
            FAIL_RETURN;
        }
        if (!term3.expr.is_num()) {
            CORAL_error(COR_BAD_ARGLIST, "Port must be a number","SocketRelSolver");
            term3.printon(Cor_exEnv->err_strm);
            *(Cor_exEnv->err_strm) << endl;
            FAIL_RETURN;
        }
        if (!term4.expr.is_symbol()) {
            CORAL_error(COR_BAD_ARGLIST, "Workspace must be a symbol","SocketRelSolver");
            term4.printon(Cor_exEnv->err_strm);
            *(Cor_exEnv->err_strm) << endl;
            FAIL_RETURN;
        }
        if (!term5.expr.is_symbol()) {
            CORAL_error(COR_BAD_ARGLIST, "Relation name must be a symbol","SocketRelSolver");
            term5.printon(Cor_exEnv->err_strm);
            *(Cor_exEnv->err_strm) << endl;
            FAIL_RETURN;
        }
       
        local_name    = term1.expr.functor();
        arity         = term1.expr.arity();
        host          = term2.expr.symbol();
        port          = term3.expr.NumIval();
        work_space    = term4.expr.symbol();
        relation_name = term5.expr.symbol();

    }
            
#ifdef DEBUG_SOCK_REL
    *(Cor_exEnv->out_strm) << "local_name = " << local_name << endl;
    *(Cor_exEnv->out_strm) << "arity = " << arity << endl;
    *(Cor_exEnv->out_strm) << "host = " << host << endl;
    *(Cor_exEnv->out_strm) << "port  = " << port << endl;
    *(Cor_exEnv->out_strm) << "work_space = " << work_space << endl;
    *(Cor_exEnv->out_strm) << "relation_name = " << relation_name << endl;
#endif
        
    Cor_Relation *new_rel;
    new_rel = Cor_find_external_relation(Cor_CArg(local_name),arity);
    if (new_rel) 
        printf("Error: there is already local relation named %s\n",local_name);
    else {
        Cor_SockRelation * sock_rel = new Cor_SockRelation(local_name, arity, host, port, work_space, relation_name);
        if (sock_rel->Valid()) {
            Cor_CurDB->RelationTable->insert(Cor_CArg(local_name), sock_rel);
            printf("socket relation %s created.\n", local_name);
        }
        else
            printf("socket relation %s cannot be created.\n", local_name);
    }
#else
    *(Cor_exEnv->err_strm)
        << "The command '" << *Cor_SocketRelSymbol << "' is unavailable.\n"
        << "Sockets are not compiled into this version of Coral.\n";

    iterator.set_no_match();
#endif // SOCK_REL

    return iterator.bindenv;
}

// *************************************************************************
