// *************************************************************************
//    TupleIterator.h
// *************************************************************************

#ifndef _TupleIterator_H_
#define _TupleIterator_H_

#include "ArgList.h"
#include "BindEnv.h"
#include "RMark.h"
#include "StackMark.h"
#include "PipelinedExecInfo.h"
#include "GenericIndex.h"
#include "ModuleInfo.h"
#include "ModuleExecInfo.h"
#include "Debug.h"

// *************************************************************************
//    Codependent classes

class Cor_PipelinedExecInfo;
class Cor_Relation;
class Cor_QFModuleData;
class Cor_RelationLink;
class Cor_TupleNode;
class Cor_TupleStack;

  // To indicate which field of the union is being used
enum Cor_IterType {
    Invalid_Iter,        // None defined
    Answer_Iter          // answer_iterator
};

// *************************************************************************

class Cor_TupleIterator {

  public:
    // Fields that are (usually) invariant within a rule application
    Cor_ArgList       arg_list;
    Cor_Unify_Funct  *arg_unify;    // Unification algthm to use for each arg

    Cor_ArgList      *del_arglist;      // Delete this in destructor
    Cor_BindEnv      *bindenv;
    Cor_Relation     *relation;

    Cor_RMark        *start_mark;

    union {
        Cor_RMark        *end_mark;
        Cor_QFModuleData *moduledata;    // For derived relations
    };

    Cor_StackMark     stack_mark;
    Cor_BindEnv      *tuple_env;

    Cor_IterType      iter_type;    // Specifies which of the below is
                                    // being used (sometimes)

    union {
        Cor_RMark *mpos;
        int       ppos;             // position of next tuple to be considered
        void      *scanDesc;        // for persistent relations
        void      *cursor;          // for persistent relations
        Cor_TupleIterator *next;    // For free store handling
        Cor_TupleIterator *answer_iterator;
        Cor_TupleNode     *SortNext; // For sorted relation
    };

    union {
        Cor_TupleNode     *SortPrev;  // For sorted relation only
        int ipos;                     // the state the iterator is in.
        Cor_PipelinedExecInfo *pipe;  // for pipelined relations only. 
    };

    Cor_TupleStack   *bkt;          // Current bucket being scanned

    int pindex;                     // index of current bucket, if the
                                    // relation is a hash relation
    int non_mt_indx;                // Current index of non_mt stack
                                    // being examined

    int _no_match;                  // To return status on calls to
                                    // get_next_tuple

    Cor_RelationLink *delta_rel_pos; // Used to handle semi-naive delta relns

    // The following three fields are for dealing with indices
    HashVal hash;
    Cor_GenericIndex *index;
    int *indexnump;

    // The following field is used by return unify.  It is set by
    // unify_literal, to signal that apply-rule should do the actual 
    // unification, since extra information about the supplementary and 
    // answer facts is required.
    int unification_number;  

    union {
        Cor_PipelinedExecInfo *calling_pei;    // used by pipelining for Cut
        Cor_Tuple *prev_tuple;
    };

    Cor_BindEnv *prev_bindenv;

    // the following field is used to store intermediate state of
    // bottom-up computation, when answers are being returned at
    // the end of each iteration
    union {
        Cor_ModuleExecInfo *state;
        Cor_ModuleInfo *prev_module_info;   // for pipelined relations only
    };
    
    // userData is meant to be used by users to store data for
    // builtins between calls  ---- wgr 5/12/93
 
    void *userData;

    void reset_fields(Cor_Relation *rel, Cor_BindEnv *env,
            Cor_RMark *start, Cor_RMark *end, int *tryindexp)
    {
        relation   = rel;
        bindenv    = env;
        start_mark = start;
        end_mark   = end;
        indexnump  = tryindexp;
    }

    Cor_TupleIterator(Cor_Relation *rel, Cor_ArgList& args,
        Cor_BindEnv *env, Cor_Unify_Funct *unify=NULL,
        Cor_RMark *start = NULL, Cor_RMark *end = NULL, int *tryindexp = NULL);

    Cor_TupleIterator(Cor_Relation *rel, Cor_Tuple *tuple,
        Cor_Unify_Funct *unify=NULL,
        Cor_RMark *start = NULL, Cor_RMark *end = NULL);
   ~Cor_TupleIterator();

    inline void reset() {
#ifdef    DO_CALL_STACK
        CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
        // Reset only those fields that change within a rule application
        tuple_env = NULL;
        ppos = 0;
        ipos = 0;
        _no_match = 0;
        delta_rel_pos = NULL;
        hash = VarHashValue;
        index = NULL;
        unification_number = -1;
        prev_tuple = NULL;
        prev_bindenv = NULL;
        state = NULL;
    }

    inline int in_delta()        { return (delta_rel_pos != NULL); }
    inline void release() {
#ifdef    DO_CALL_STACK
        CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif

        // Uncomment the following to break pipelined evaluation. -SPF
        // stack_mark.pop_to();
        if (relation) {
            (void) relation->scan_count--;
            if (relation->has_release)
                relation->release(*this);
        }
    }

    inline int no_match()        { return _no_match;}
    inline void set_no_match()   { _no_match = 1;}
    inline void reset_no_match() {_no_match = 0;}
    Cor_BindEnv *get_next();

    // get_next_tuple(): returns tuple if there is one (Builtin rels etc
    // return NULL). 
    inline Cor_Tuple *get_next_tuple() {
        return relation->get_next_tuple(*this);
    }
};

// *************************************************************************

#endif _TupleIterator_H_
