// *************************************************************************
//    SipGen.C
// *************************************************************************

#include "SipGen.h"

#include "Config.h"
#include "Collection.h"
#include "ClauseInfo.h"
#include "Literal.h"
#include "Sips.h"
#include "Globals.h"

// *************************************************************************

#ifdef    NOT_USED
static void get_adorn(Cor_Clause *query, Cor_BitVector& adorn);
#endif

static int setup_head_arg(Cor_CArg *arg, void *data);
static void setup_head_vars(ClauseInfo *rule_info);
static int gen_sip_var(Cor_CArg *arg, void *data);
static int sip_exists(Cor_Clause *rule, Cor_BitVector& adorn, int pred_num);

// *************************************************************************
//  Creates sips for each rule (including the query) of the program, for
//  each possible adornment.  The sip for the query is generated only for
//  the specific bindings of the query predicate.  

void create_sips()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_BitVector adornment;
    int num_args;
    long j;
    ClauseInfo rule_info[1];
    extern Cor_Collection *pre_ad_rules;
    Cor_Clause *rule;
    int k, nrules;

    nrules = pre_ad_rules->count();

    for (k=0; k < nrules; k++) {
        rule = pre_ad_rules->get_ith(k);

        num_args = rule->head->args.count();
        if (num_args > COR_MAX_ARGLIST_SIZE) {
            cerr_strm->printf(
                "Rule (line %d) has too many args (%d) - ignored\n",
                rule->line_number, num_args);
            cerr_strm->flush();
            continue;
        }
        setup_vars(rule, rule_info);
        // WARNING: This code will only work if there are less than
        // 32 arguments!!  REWRITE IT
        for (j = 0; j < (1 << num_args); j++) {
            Cor_BitVector adorn(num_args, j);
            gen_sip_rule(rule_info, adornment);
        }
        cleanup_vars(rule_info);
    }
}

#ifdef    NOT_USED
// *************************************************************************
//  Gets the adornment corresponding to the query predicate.  Any argument
//  that is a constant is considered bound, and all others are free. 
static void get_adorn(Cor_Clause *query, Cor_BitVector& adorn)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    adorn.setlen(query->head->args.count());

    int i=0;
    FOR_EACH_ARG(arg, query->head->args) {
        if (arg->isConstant()) {
            adorn.set(i);
            i++;
        }
    } END_EACH_ARG
}

#endif
// *************************************************************************
//  Helper function called (by setup_vars) on each variable to set
//  var_defining_pred to the first predicate where the variable occurs.
int setup_arg(Cor_CArg *arg, void *data)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    ClauseInfo *rule_info = (ClauseInfo*)data;

    if (rule_info->var_defining_pred[arg->var_num()] == -1)
        rule_info->var_defining_pred[arg->var_num()] = rule_info->i;
    return 0;
}

// *************************************************************************
//  Sets up data structures (in rule_info) to deal with variables:
//    var_head_arg_pos - set up by calling setup_head_vars(rule).
//    var_defining_pred - set up by going through each argument of each 
//        body predicate and using scan_vars to call
//        setup_arg on each variable.

void setup_vars(Cor_Clause *rule, ClauseInfo *rule_info)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int i;
    int num_args;
    int rule_cnt;

    Cor_BitVector *cur_var_bits;
    rule_info->rule = rule;

    if (rule->var_names)
        rule_cnt = rule->var_names->count();
    else
        rule_cnt = 0;

    rule_info->var_defining_pred = new int[rule_cnt+1];
    rule_info->var_head_arg_pos = cur_var_bits =
        new Cor_BitVector[rule_cnt+1];

    num_args = rule->head->args.count();
    for (i = 0; i < rule_cnt; i++, cur_var_bits++) {
        rule_info->var_defining_pred[i] = -1;
        cur_var_bits->setlen(num_args);
    }
    setup_head_vars(rule_info);
    for (i = 0; i < rule->num_preds; i++) {
        FOR_EACH_ARG(arg, rule->preds[i]->args) {
            //    Goes though an argument ( of the i'th body predicate, and for
            //    all variables that occur in it, it sets the var_defining_pred
            //    to i, if it has not already been set.  Thus var_defining_pred
            //    is set to the first predicate where the variable occurs.
            rule_info->i = i;
            arg->scan_vars(rule_info, setup_arg);
        } END_EACH_ARG
    }
}

// *************************************************************************
//  Sets the corresoponding bit of the variable position in var_head_arg_pos
//  to indicate that the variable occurs in that position.
static int setup_head_arg(Cor_CArg *arg, void *data)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    ClauseInfo *rule_info = (ClauseInfo*)data;
    rule_info->var_head_arg_pos[arg->var_num()].set(rule_info->i);
    return(0);
}

// *************************************************************************
//  Goes through the head of a rule, and for each variable, sets the bits in
//  var_head_arg_pos to indicate which argument positions in the head the
//  variable occurs in.  This is used later to check if the variable is bound. 

static void setup_head_vars(ClauseInfo *rule_info)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int i = 0;

    FOR_EACH_ARG(arg, rule_info->rule->head->args) {
        rule_info->i = i++;
        arg->scan_vars(rule_info, setup_head_arg);
    } END_EACH_ARG
}

// *************************************************************************
void cleanup_vars(ClauseInfo *rule_info)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    delete  rule_info->var_defining_pred;
    //delete  rule_info->var_head_arg_pos;
    // WHY WAS THIS CAUSING A MEMORY OVERWRITE ????- PRAVEEN
}

// *************************************************************************
//  ??? The following function (and its calls) can be cleaned up to make 
//  it more efficient in generating sips.

static int gen_sip_var(Cor_CArg *arg, void *data)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    ClauseInfo *rule_info = (ClauseInfo*) data;

    if (rule_info->sip_generated[arg->var_num()])
        return 0;

    if ((rule_info->var_head_arg_pos[arg->var_num()] & *rule_info->adorn)
        || ((rule_info->var_defining_pred[arg->var_num()] < rule_info->i)
        && (rule_info->var_defining_pred[arg->var_num()] != -1)))
    {
        // We have checked that the variable was defined in the head and
        // that it occurs in a bound position of the adornment  OR the
        // defining predicate of this variable occurs earlier in the
        // body. In either case the variable is bound and has a sip arc
        // entering the predicate occurrence with this variable as a label.

        Cor_Sip *sip = new Cor_Sip();
        sip->tail = *rule_info->tail;
        sip->head = rule_info->i + 1;    // 0 indicates head of rule
        sip->bound_vars.set(arg->var_num());

        // Linking it up to the rule.
        sip->next = rule_info->cur_sips->sip_list;
        rule_info->cur_sips->sip_list = sip;

        // Have generated sip arc for this variable.
        rule_info->sip_generated[arg->var_num()] = 1;
    }
    return 0;
}

// *************************************************************************
//  This function fills in sips for the specified rule and adornment,
//  for each body literal for which the user has not specified sips.
//  It generates sips according to the left-right ordering.

void gen_sip_rule(ClauseInfo *rule_info, Cor_BitVector& adorn)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_Sips *sips;
    char *sip_generated;
    int i, j;
    Cor_Clause *rule = rule_info->rule;
    int rule_cnt;

    // Check if there is sips node for this adorn.
    for (sips=rule->sips;sips;sips=sips->next) {
        if (sips->adorn == adorn)
            break;
    }

    if (!sips) {
        // Create a new sips node and link it up to the rule node.
        sips = new Cor_Sips();
        sips->next = rule->sips;
        rule->sips = sips;
        sips->adorn = adorn;
        sips->sip_list = NULL;
    }

    // Has a sip been generated for this var in this predicate occurrence?
    if (rule->var_names)
        rule_cnt = rule->var_names->count();
    else
        rule_cnt = 0;

    sip_generated = new char [rule_cnt+1];

    Cor_BitVector tail(rule->num_preds + 1);

    for (i = 0; i < rule->num_preds; i++) {
        // This is a total sip in that every predicate occurrence to the left
        // of a predicate occurrence (including the head) is part of the tail
        // of every sip arc entering this predicate occurrence.  Bit 0
        // corresponds to the head and bit i corresponds to the i'th body
        // predicate (with index i-1 in preds[]).

        tail.set(i);

        if (sip_exists(rule, adorn,i))
            continue;

        for (j = 0; j < rule_cnt; j++) {

            // In case of multiple occurrences of the same variable in a
            // predicate occurrence, we desire to generate only one sip-arc.
            sip_generated[j] = 0;
        }

        FOR_EACH_ARG(arg, rule->preds[i]->args) {
            rule_info->sip_generated = sip_generated;
            rule_info->i = i;
            rule_info->adorn = &adorn;
            rule_info->tail = &tail;
            rule_info->cur_sips=sips;
            arg->scan_vars(rule_info, gen_sip_var);
            if (! rule_info->cur_sips->sip_list) {
                // None of the variables was bound.  Create a sip anyway
                // with no variable bound.
                
                Cor_Sip *sip = new Cor_Sip();
                sip->tail = *rule_info->tail;
                sip->head = i + 1;        // 0 indicates head of rule

                // Linking it up to the rule.
                sip->next = rule_info->cur_sips->sip_list;
                rule_info->cur_sips->sip_list = sip;
            }
        } END_EACH_ARG
    }
    delete [] sip_generated;
}

// *************************************************************************
static int sip_exists(Cor_Clause *rule, Cor_BitVector& adorn, int pred_num)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_Sips *sips;
    Cor_Sip *sip;

    for (sips = rule->sips; sips != NULL; sips = sips->next) {
        if (adorn == sips->adorn) {
           for (sip=sips->sip_list;sip;sip=sip->next) {
                // pred_num+1 is used since in the sip structure an index
                // of 0 means the head, rather than the 0'th body predicate.
                // Since a sip with the same adornment and head exists we
                // don't generate sip arcs corresponding to this adornment.
                if (sip->head == pred_num+1)
                    return 1;
            }
        }
    }
    return (0);
}

// *************************************************************************
